local VERSION = tonumber(("$Revision: 32678 $"):match("%d+"))

local PitBull = PitBull
local PitBull_Combat = PitBull:NewModule("Combat", "AceEvent-2.0")
local self = PitBull_Combat
if PitBull.revision < VERSION then
	PitBull.version = "r" .. VERSION
	PitBull.revision = VERSION
	PitBull.date = ("$Date: 2007-04-15 01:47:10 -0400 (Sun, 15 Apr 2007) $"):match("%d%d%d%d%-%d%d%-%d%d")
end

-- localization placeholder
local L = setmetatable({}, {__index = function(self, key) self[key] = key; return key end })

local new, del, newHash = PitBull.new, PitBull.del, PitBull.newHash
local newFrame, delFrame = PitBull.newFrame, PitBull.delFrame

PitBull_Combat:RegisterPitBullChildFrames('combatIcon')

function PitBull_Combat:OnInitialize()
	self.db = PitBull:AcquireDBNamespace("Combat")
	PitBull:RegisterDefaults("Combat", "profile", {
		groups = {
			['**'] = { disable = true },
			["player"] = {
				disable = false,
			},			
			["target"] = {
				disable = false,
			},			
			["party"] = {
				disable = false,
			},			
		},
	})
end

function PitBull_Combat:OnEnable()
	self:RegisterEvent("PLAYER_REGEN_ENABLED")
	self:RegisterEvent("PLAYER_REGEN_DISABLED", "PLAYER_REGEN_ENABLED")
	
	self:ScheduleRepeatingEvent("PitBull_Combat-Update", self.UpdateAllNonWacky, 0.5, self)
end

function PitBull_Combat:PLAYER_REGEN_ENABLED()
	local frame = PitBull:GetUnitFrame('player')
	if frame then
		self:Update('player', frame)
	end
end

function PitBull_Combat:Update(unit, frame)
	if self.db.profile.groups[frame.group].disable then return end

	local inCombat = UnitAffectingCombat(unit)
	if PitBull.configMode and not UnitExists(unit) then
		inCombat = true
	end
	
	if inCombat then
		if not frame.combatIcon then
			frame.combatIcon = newFrame("Texture", frame.overlay, "ARTWORK")
			frame.combatIcon:SetTexture("Interface\\CharacterFrame\\UI-StateIcon")
			frame.combatIcon:SetTexCoord(0.5, 1, 0, 0.5)
			PitBull:UpdateLayout(frame)
		end
	else
		if frame.combatIcon then
			frame.combatIcon = delFrame(frame.combatIcon)
			
			PitBull:UpdateLayout(frame)
		end
	end
end

function PitBull_Combat:UpdateAllNonWacky()
	for unit, frame in PitBull:IterateNonWackyUnitFrames() do
		self:Update(unit, frame)
	end
end

function PitBull_Combat:OnUpdateFrame(unit, frame)
	self:Update(unit, frame)
end

function PitBull_Combat:OnClearUnitFrame(unit, frame)
	if frame.combatIcon then
		frame.combatIcon = delFrame(frame.combatIcon)
	end
end

function PitBull_Combat:OnUnknownLayout(unit, frame)
	frame.combatIcon:SetPoint("CENTER", frame, "BOTTOMLEFT", 0, 0)
	frame.combatIcon:SetWidth(18)
	frame.combatIcon:SetHeight(18)
end

-- Unit/Group Specific Options
local function getDisabled(group)
	return PitBull_Combat.db.profile.groups[group].disable
end
local function setDisabled(group, value)
	PitBull_Combat.db.profile.groups[group].disable = value
	if value then
		for unit,frame in PitBull:IterateUnitFramesByGroup(group) do
			if frame.combatIcon then
				frame.combatIcon = delFrame(frame.combatIcon)
				PitBull:UpdateLayout(frame)
			end
		end
	else
		for unit,frame in PitBull:IterateUnitFramesByGroup(group) do
			if not frame.combatIcon then
				self:Update(unit, frame)
			end
		end
	end
end

PitBull_Combat:RegisterPitBullOptionsMethod(function(group)
	return {
		name = L["Combat Icon"],
		desc = L["Combat Icon"],
		type = 'group',
		args = {
			ignore = {
				type = 'toggle',
				name = L["Disable Combat Icon"],
				desc = L["Disable Combat Icon"],
				get = getDisabled,
				set = setDisabled,
				passValue = group,
			},
		}
	}
end)
