local VERSION = tonumber(('$Revision: 32881 $'):match('%d+'))

local PitBull = PitBull
if PitBull.revision < VERSION then
	PitBull.version = 'r' .. VERSION
	PitBull.revision = VERSION
	PitBull.date = ('$Date: 2007-04-17 04:51:05 -0400 (Tue, 17 Apr 2007) $'):match('%d%d%d%d%-%d%d%-%d%d')
end

local ABF_db = PitBull:AcquireDBNamespace('Layout-ABF')
PitBull:RegisterDefaults('Layout-ABF', 'profile', {
	['**'] = {
		portraitSide = 'left',
		auraSide = 'bottom',
		--note, these are only used when debuffs are split
		buffSide = 'bottom',
		debuffSide = 'right',
	},
	target = {
		portraitSide = 'right',
		auraSide = 'right',
		comboSide = 'bottom',
		comboTexture = 'default',
		comboText = false,
	}
})

local function portraitSide_get(group)
	return ABF_db.profile[group].portraitSide
end
local function portraitSide_set(group, value)
	ABF_db.profile[group].portraitSide = value
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		PitBull:UpdateLayout(frame)
	end
end

local function auraSide_get(group)
	return ABF_db.profile[group].auraSide
end
local function auraSide_set(group, value)
	ABF_db.profile[group].auraSide = value
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		PitBull:UpdateLayout(frame)
	end
end

local function buffSide_get(group)
	return ABF_db.profile[group].buffSide
end
local function buffSide_set(group, value)
	ABF_db.profile[group].buffSide = value
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		PitBull:UpdateLayout(frame)
	end
end

local function debuffSide_get(group)
	return ABF_db.profile[group].debuffSide
end
local function debuffSide_set(group, value)
	ABF_db.profile[group].debuffSide = value
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		PitBull:UpdateLayout(frame)
	end
end

local function splitOptions_isHidden(group)
	return not PitBull:GetModule('Aura').db.profile[group].split
end

local function nonsplitOptions_isHidden(group)
	return PitBull:GetModule('Aura').db.profile[group].split
end

local function comboSide_get(group)
	return ABF_db.profile[group].comboSide
end
local function comboSide_set(group, value)
	ABF_db.profile[group].comboSide = value
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		PitBull:UpdateLayout(frame)
	end
end

local function comboTexture_get(group)
	return ABF_db.profile[group].comboTexture
end
local function comboTexture_set(group, value)
	ABF_db.profile[group].comboTexture = value
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		PitBull:UpdateLayout(frame)
	end
end

local function comboText_get(group)
	return ABF_db.profile[group].comboText
end
local function comboText_set(group, value)
	ABF_db.profile[group].comboText = value
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		PitBull:UpdateLayout(frame)
	end
end

PitBull:RegisterLayout('ABF', {
	healthBar = function(unit, frame)
		local oneunit = (frame:GetHeight()) / 8
		local height = oneunit * 3 - 1
		if frame.manaBar and not frame.castBar then
			height = oneunit * 4 - 1
		elseif not frame.manaBar and frame.castBar then
			height = oneunit * 6 - 1
		elseif not frame.manaBar and not frame.castBar then
			height = oneunit * 8
		end
		frame.healthBar:SetPoint("TOP", frame, "TOP", 0, 0)
		if frame.portrait then
			if ABF_db.profile[frame.group].portraitSide == "right" then
				frame.healthBar:SetPoint("LEFT", frame, "LEFT", 0, 0)
				frame.healthBar:SetPoint("RIGHT", frame.portrait, "LEFT", -2, 0)
			else
				frame.healthBar:SetPoint("LEFT", frame.portrait, "RIGHT", 2, 0)
				frame.healthBar:SetPoint("RIGHT", frame, "RIGHT", 0, 0)
			end
		else
			frame.healthBar:SetPoint("LEFT", frame, "LEFT", 0, 0)
			frame.healthBar:SetPoint("RIGHT", frame, "RIGHT", 0, 0)
		end
		frame.healthBar:SetHeight(height)		
	end,
	manaBar = function(unit, frame)
		local oneunit = (frame:GetHeight()) / 8
		local height = oneunit * 3 - 1
		if frame.healthBar and not frame.castBar then
			height = oneunit * 4 - 1
		end
		if not frame.healthBar and not frame.castBar then
			height = oneunit * 8
		end
		if frame.healthBar then
			frame.manaBar:SetPoint("TOP", frame.healthBar, "BOTTOM", 0, -1)
		else
			frame.manaBar:SetPoint("TOP", frame, "TOP", 0, 0)
		end
		if frame.portrait then
			if ABF_db.profile[frame.group].portraitSide == "right" then
				frame.manaBar:SetPoint("LEFT", frame, "LEFT", 0, 0)
				frame.manaBar:SetPoint("RIGHT", frame.portrait, "LEFT", -2, 0)
			else
				frame.manaBar:SetPoint("LEFT", frame.portrait, "RIGHT", 2, 0)
				frame.manaBar:SetPoint("RIGHT", frame, "RIGHT", 0, 0)
			end
		else
			frame.manaBar:SetPoint("LEFT", frame, "LEFT", 0, 0)
			frame.manaBar:SetPoint("RIGHT", frame, "RIGHT", 0, 0)
		end
		frame.manaBar:SetHeight(height)	
	end,
	castBar = function(unit, frame)
		local oneunit = frame:GetHeight() / 8
		local height = oneunit * 2
		frame.castBar:SetPoint("BOTTOM", frame, "BOTTOM", 0, 0)
		frame.castBar:SetPoint("LEFT", frame, "LEFT", height, 0)
		frame.castBar:SetPoint("RIGHT", frame, "RIGHT", 0, 0)
		frame.castBar:SetHeight( height )
--		frame.castBar:Hide()
		frame.castBar.bg:ClearAllPoints()
		frame.castBar.bg:SetPoint("LEFT", frame, "LEFT", 0, 0 )
		frame.castBar.bg:SetPoint("BOTTOM", frame, "BOTTOM", 0, 0 )
		frame.castBar.bg:SetPoint("RIGHT", frame, "RIGHT", 0, 0 )
		frame.castBar.bg:SetHeight( height )
		frame.castBar.bg:Show()
	end,
	castBarIcon = function(unit, frame )
		local oneunit = (frame:GetHeight()) / 8
		local height = oneunit * 2
		frame.castBarIcon:SetHeight( height )
		frame.castBarIcon:SetWidth( height )
		frame.castBarIcon:SetPoint("RIGHT", frame.castBar, "LEFT" )
		frame.castBarIcon:SetTexCoord(0.07, 0.93, 0.07, 0.93)
		frame.castBarIcon:Hide()
	end,
	castBarText = function(unit, frame)
		frame.castBarText:ClearAllPoints()
		frame.castBarText:SetAllPoints(frame.castBar)
	end,
	castBarTimeText = function(unit, frame)
		frame.castBarTimeText:ClearAllPoints()
		frame.castBarTimeText:SetAllPoints(frame.castBar)
		frame.castBarTimeText:SetJustifyH("RIGHT")
	end,
	healthText = function(unit, frame)
		if frame.healthBar then
			frame.healthText:SetPoint('RIGHT', frame.healthBar, 'RIGHT', -3, 0)
		end
	end,
	manaText = function(unit, frame)
		if frame.manaBar then
			frame.manaText:SetPoint('RIGHT', frame.manaBar, 'RIGHT', -3, 0)
		end	
	end,
	nameText = function(unit, frame)
		if frame.healthBar then
			frame.nameText:SetPoint('LEFT', frame.healthBar, 'LEFT', 3, 0)
			if frame.healthText then
				frame.nameText:SetPoint('RIGHT', frame.healthText, 'LEFT', 0, 0)
			end
		end
	end,
	classText = function(unit, frame)
		if frame.manaBar then
			frame.classText:SetPoint('LEFT', frame.manaBar, 'LEFT', 3, 0)
			frame.classText:SetJustifyV("MIDDLE")
			if frame.manaText then
				frame.classText:SetPoint('RIGHT', frame.manaText, 'LEFT', 0, 0)
			end
		end
	end,
	auraFrame = function(unit, frame)
		frame.auraFrame:ClearAllPoints()
		local size = PitBull:GetModule('Aura').db.profile[frame.group].buffsize
		frame.auraFrame:SetHeight(size)
		if PitBull:GetModule('Aura').db.profile[frame.group].split then
			local buffside = ABF_db.profile[frame.group].buffSide
			local w = frame:GetWidth()
			frame.auraFrame:SetWidth(w)
			if buffside == 'bottom' then
				frame.auraFrame:SetPoint('TOPLEFT', frame, 'BOTTOMLEFT', -2, -2)
			elseif buffside == 'top' then
				frame.auraFrame:SetPoint('BOTTOMLEFT', frame, 'TOPLEFT', -2, 2)
			elseif buffside == 'left' then
				frame.auraFrame:SetPoint('BOTTOMRIGHT', frame, 'LEFT', w-size-2, -2)
			elseif buffside == 'right' then
				frame.auraFrame:SetPoint('BOTTOMLEFT', frame, 'RIGHT', 2, 0)
			elseif buffside == 'topleft' then
				frame.auraFrame:SetPoint('TOPRIGHT', frame, 'TOPLEFT', w-size-2, 0)
			elseif buffside == 'topright' then
				frame.auraFrame:SetPoint('TOPLEFT', frame, 'TOPRIGHT', 2, 0)
			elseif buffside == 'bottomleft' then
				frame.auraFrame:SetPoint('BOTTOMRIGHT', frame, 'BOTTOMLEFT', w-size-2, 0)
			elseif buffside == 'bottomright' then
				frame.auraFrame:SetPoint('BOTTOMLEFT', frame, 'BOTTOMRIGHT', 2, 0)
			end
		else
			local side = ABF_db.profile[frame.group].auraSide
			if side == 'bottom' then
				frame.auraFrame:SetPoint('TOPLEFT', frame, 'BOTTOMLEFT', -2, -2)
				frame.auraFrame:SetPoint('TOPRIGHT', frame, 'BOTTOMRIGHT', 2, -2)
			elseif side == 'top' then
				local split = PitBull:GetModule('Aura').db.profile[frame.group].split
				frame.auraFrame:SetPoint('BOTTOMLEFT', frame, 'TOPLEFT', -2, split and 18 or 2)
				frame.auraFrame:SetPoint('BOTTOMRIGHT', frame, 'TOPRIGHT', 2, split and 18 or 2)
			elseif side == 'left' then
				frame.auraFrame:SetWidth(160)
				frame.auraFrame:SetPoint('BOTTOMRIGHT', frame, 'LEFT', -2, 0)
			elseif side == 'topright' then
				frame.auraFrame:SetWidth(160)
				frame.auraFrame:SetPoint('TOPLEFT', frame, 'TOPRIGHT', 2, 0)
			elseif side == 'bottomright' then
				frame.auraFrame:SetWidth(160)
				frame.auraFrame:SetPoint('BOTTOMLEFT', frame, 'BOTTOMRIGHT', 2, 0)
			elseif side == 'bottomleft' then
				frame.auraFrame:SetWidth(160)
				frame.auraFrame:SetPoint('BOTTOMRIGHT', frame, 'BOTTOMLEFT', -2, 0)
			elseif side == 'topleft' then
				frame.auraFrame:SetWidth(160)
				frame.auraFrame:SetPoint('TOPRIGHT', frame, 'TOPLEFT', -2, 0)
			elseif side == 'right' then
				frame.auraFrame:SetWidth(160)
				frame.auraFrame:SetPoint('BOTTOMLEFT', frame, 'RIGHT', 2, 0)
			end
		end
	end,
	auraFrame2 = function(unit,frame)
		frame.auraFrame2:ClearAllPoints()
		local size = PitBull:GetModule('Aura').db.profile[frame.group].debuffsize
		local w = frame:GetWidth()
		frame.auraFrame2:SetHeight(size)
		frame.auraFrame2:SetWidth(w)
		local debuffside = ABF_db.profile[frame.group].debuffSide
		if debuffside == 'bottom' then
			frame.auraFrame2:SetPoint('TOPLEFT', frame, 'BOTTOMLEFT', -2, -2)
		elseif debuffside == 'top' then
			frame.auraFrame2:SetPoint('BOTTOMLEFT', frame, 'TOPLEFT', -2, 2)
		elseif debuffside == 'left' then
			frame.auraFrame2:SetPoint('BOTTOMRIGHT', frame, 'LEFT', w-size-2, -2)
		elseif debuffside == 'right' then
			frame.auraFrame2:SetPoint('BOTTOMLEFT', frame, 'RIGHT', 2, 0)
		elseif debuffside == 'topleft' then
			frame.auraFrame2:SetPoint('TOPRIGHT', frame, 'TOPLEFT', w-size-2, 0)
		elseif debuffside == 'topright' then
			frame.auraFrame2:SetPoint('TOPLEFT', frame, 'TOPRIGHT', 2, 0)
		elseif debuffside == 'bottomleft' then
			frame.auraFrame2:SetPoint('BOTTOMRIGHT', frame, 'BOTTOMLEFT', w-size-2, 0)
		elseif debuffside == 'bottomright' then
			frame.auraFrame2:SetPoint('BOTTOMLEFT', frame, 'BOTTOMRIGHT', 2, 0)
		end
	end,
	portrait = function(unit, frame)
		if ABF_db.profile[frame.group].portraitSide ~= 'right' then
			frame.portrait:SetPoint('TOPLEFT', frame, 'TOPLEFT', 0, 0)
		else
			frame.portrait:SetPoint('TOPRIGHT', frame, 'TOPRIGHT', 0, 0)
		end
		local oneunit = (frame:GetHeight()) / 8
		local size = oneunit * 8 - 1
		if frame.castBar then
			size = oneunit * 6 - 1
		end
		frame.portrait:SetWidth( size )
		frame.portrait:SetHeight( size )
	end,
	combatText = function(unit, frame)
		if frame.portrait then
			frame.combatText:SetPoint("CENTER", frame.portrait, "CENTER", 0, 0)
		else
			frame.combatText:SetPoint("CENTER", frame, "CENTER", 0, 0)
		end
	end,
	comboPoints = function(unit, frame)
		if frame.group ~= 'target' then return end 
		local side = ABF_db.profile[frame.group].comboSide
		if side == 'left' or side == 'right' then
			if side == 'left' then
				frame.comboPoints:SetPoint( 'RIGHT', frame, 'LEFT', 0, 0)
			else
				frame.comboPoints:SetPoint( 'LEFT', frame, 'RIGHT', 0, 0)
			end
			frame.comboPoints:SetWidth( 10 )
			frame.comboPoints:SetHeight( 50 )
			for i = 1, 5 do
				local combo = frame.comboPoints['combo'..i]
				if ABF_db.profile[frame.group].comboTexture == 'default' then
					combo:SetTexture('Interface\\AddOns\\PitBull\\ComboPoints\\combo.tga')
					combo:SetWidth( 10 )
					combo:SetHeight( 10 )
				else
					combo:SetTexture(PitBull:GetStatusBarTexture())
					combo:SetWidth( 5 )
					combo:SetHeight( 8 )
				end
				if i == 1 then
					combo:SetPoint('BOTTOM', frame.comboPoints, 'BOTTOM')
				else
					combo:SetPoint('BOTTOM', frame.comboPoints['combo'..(i-1)], 'TOP', 0, 2)
				end			
			end
		elseif side == 'top' or side == 'bottom' then
			if side == 'top' then
				frame.comboPoints:SetPoint( 'TOPRIGHT', frame, 'TOPRIGHT', 0, 0)
			else
				frame.comboPoints:SetPoint( 'BOTTOMRIGHT', frame, 'BOTTOMRIGHT', 0, 0)
			end
			frame.comboPoints:SetWidth( 50 )
			frame.comboPoints:SetHeight( 10 )
			for i = 1, 5 do
				local combo = frame.comboPoints['combo'..i]
				if ABF_db.profile[frame.group].comboTexture == 'default' then
					combo:SetTexture('Interface\\AddOns\\PitBull\\ComboPoints\\combo.tga')
					combo:SetWidth( 10 )
					combo:SetHeight( 10 )
				else
					combo:SetTexture(PitBull:GetStatusBarTexture())
					combo:SetWidth( 5 )
					combo:SetHeight( 8 )
				end
				if i == 1 then
					combo:SetPoint('RIGHT', frame.comboPoints, 'RIGHT')
				else
					combo:SetPoint('RIGHT', frame.comboPoints['combo'..(i-1)], 'LEFT')
				end	
			end
		end
	end,
	comboText = function(unit, frame)
		if ABF_db.profile[frame.group].comboText or not frame.manaBar then
			frame.comboText:Hide()
		else
			frame.comboText:SetPoint('CENTER', frame.manaBar, 'CENTER')
		end
	end,
	combatIcon = function(unit, frame)
		frame.combatIcon:SetPoint("CENTER", frame, "BOTTOMLEFT", 0, 0)
		frame.combatIcon:SetWidth(24)
		frame.combatIcon:SetHeight(24)
	end,
	pvpIcon = function(unit, frame)
		frame.pvpIcon:SetPoint("CENTER", frame, "TOPRIGHT", 0, 0)
		frame.pvpIcon:SetWidth(27)
		frame.pvpIcon:SetHeight(27)
	end,
	restIcon = function(unit, frame)
		if frame.combatIcon then
			frame.restIcon:SetPoint("LEFT", frame.combatIcon, "RIGHT", 4, 0)
		else
			frame.restIcon:SetPoint("CENTER", frame, "BOTTOMLEFT", 0, 0)
		end
		frame.restIcon:SetWidth(24)
		frame.restIcon:SetHeight(24)
	end,
	leaderIcon = function(unit, frame)
		frame.leaderIcon:SetPoint("CENTER", frame, "TOPLEFT", 0, 0)
		frame.leaderIcon:SetWidth(18)
		frame.leaderIcon:SetHeight(18)
	end,
	masterIcon = function(unit, frame)
		if frame.leaderIcon then
			frame.masterIcon:SetPoint("LEFT", frame.leaderIcon, "RIGHT", 4, 0)
		else
			frame.masterIcon:SetPoint("CENTER", frame, "TOPLEFT", 0, 0)
		end
		frame.masterIcon:SetWidth(20)
		frame.masterIcon:SetHeight(20)
	end,
	portrait_options = function(group)
		return {
			name = 'Portrait side',
			desc = 'Portrait side',
			type = 'text',
			validate = {
				left = 'Left',
				right = 'Right',
			},
			get = portraitSide_get,
			set = portraitSide_set,
			passValue = group
		}
	end,
	auraFrame_options = function(group)
		return {
			name = 'Aura Options',
			desc = 'Aura Options',
			type = 'group',
			args = {
				nonsplitside = {
					name = 'Aura side',
					desc = 'Aura side',
					type = 'text',
					validate = {
						'left',
						'topleft',
						'bottomleft',
						'right',
						'topright',
						'bottomright',
						'top',
						'bottom',
					},
					get = auraSide_get,
					set = auraSide_set,
					passValue = group,
					hidden = nonsplitOptions_isHidden,
				},
				buffside = {
					name = 'Buff side',
					desc = 'Buff side',
					type = 'text',
					validate = {
						'left',
						'topleft',
						'bottomleft',
						'right',
						'topright',
						'bottomright',
						'top',
						'bottom',
					},
					get = buffSide_get,
					set = buffSide_set,
					passValue = group,
					hidden = splitOptions_isHidden,
				},
				debuffside = {
					name = 'Debuff side',
					desc = 'Debuff side',
					type = 'text',
					validate = {
						'left',
						'topleft',
						'bottomleft',
						'right',
						'topright',
						'bottomright',
						'top',
						'bottom',
					},
					get = debuffSide_get,
					set = debuffSide_set,
					passValue = group,
					hidden = splitOptions_isHidden,
				},
			}
		}
	end,
	comboPoints_options = function(group)
		return {
			name = 'Combo Points',
			desc = 'Combo Points options',
			type = 'group',
			hidden = function() if group ~= "target" then return true end end,
			args = {
				Side = {
					name = 'Side',
					desc = 'Side',
					type = 'text',
					validate = {
						'left',
						'right',
						'top',
						'bottom',
					},
					get = comboSide_get,
					set = comboSide_set,
					passValue = group,
				},
				Texture = {
						name = 'Texture',
						desc = 'Texture',
						type = 'text',
						validate = {
							'default',
							'bar',
						},
						get = comboTexture_get,
						set = comboTexture_set,
						passValue = group,
				},
				Text = {
						name = 'Hide Text',
						desc = 'Hide Text',
						type = 'toggle',
						get = comboText_get,
						set = comboText_set,
						passValue = group,
				},
			}
		}
	end,
})
