local VERSION = tonumber(("$Revision: 32681 $"):match("%d+"))

local PitBull = PitBull
local PitBull_MasterLooter = PitBull:NewModule("MasterLooter", "AceEvent-2.0")
local self = PitBull_MasterLooter
if PitBull.revision < VERSION then
	PitBull.version = "r" .. VERSION
	PitBull.revision = VERSION
	PitBull.date = ("$Date: 2007-04-15 02:08:16 -0400 (Sun, 15 Apr 2007) $"):match("%d%d%d%d%-%d%d%-%d%d")
end

local new, del, newHash = PitBull.new, PitBull.del, PitBull.newHash
local newFrame, delFrame = PitBull.newFrame, PitBull.delFrame

PitBull_MasterLooter:RegisterPitBullChildFrames('masterIcon')

function PitBull_MasterLooter:OnInitialize()
	self.db = PitBull:AcquireDBNamespace("MasterLooter")
	PitBull:RegisterDefaults("MasterLooter", "profile", {
		['*'] = {
			disable = false,
		},
	})
end

function PitBull_MasterLooter:OnEnable()
	self:RegisterEvent("PARTY_LOOT_METHOD_CHANGED")
	self:RegisterEvent("PARTY_MEMBERS_CHANGED", "PARTY_LOOT_METHOD_CHANGED")
	self:RegisterEvent("RAID_ROSTER_UPDATE", "PARTY_LOOT_METHOD_CHANGED")
end

function PitBull_MasterLooter:Update(unit, frame)
	if frame.group ~= "player" and frame.group ~= "party" and frame.group ~= "raid" then
		return
	end
	local has = false
	if unit == "player" then
		local _, lootmaster = GetLootMethod()
		if lootmaster == 0 then
			has = true
		end
	else
		local raid = unit:match("^raid(%d%d?)$")
		if raid then
			lootMaster = select(11, GetRaidRosterInfo(tonumber(raid)))
			if lootMaster then
				has = true
			end
		else
			local _, lootmaster = GetLootMethod()
			if lootmaster and lootmaster > 0 then
				local party = unit:match("^party(%d)$")
				if party then
					if lootmaster == tonumber(party) then
						has = true
					end
				end
			end
		end
	end
	if PitBull.configMode and not UnitExists(unit) then
		has = true
	end
	if self.db.profile[frame.group].disable then
		has = false
	end

	if has then
		if not frame.masterIcon then
			local masterIcon = newFrame("Texture", frame.overlay, "ARTWORK")
			frame.masterIcon = masterIcon
			masterIcon:SetTexture("Interface\\GroupFrame\\UI-Group-MasterLooter")
		
			PitBull:UpdateLayout(frame)
		end
	else
		if frame.masterIcon then
			frame.masterIcon = delFrame(frame.masterIcon)
			
			PitBull:UpdateLayout(frame)
		end
	end
end

function PitBull_MasterLooter:PARTY_LOOT_METHOD_CHANGED()
	for unit, frame in PitBull:IterateNonWackyUnitFrames() do
		self:Update(unit, frame)
	end
end


function PitBull_MasterLooter:OnUpdateFrame(unit, frame)
	self:Update(unit, frame)
end

function PitBull_MasterLooter:OnClearUnitFrame(unit, frame)
	if frame.masterIcon then
		frame.masterIcon = delFrame(frame.masterIcon)
	end
end

function PitBull_MasterLooter:OnUnknownLayout(unit, frame)
	if frame.leaderIcon then
		frame.masterIcon:SetPoint("LEFT", frame.leaderIcon, "RIGHT", 4, 0)
	else
		frame.masterIcon:SetPoint("CENTER", frame, "TOPLEFT", 0, 0)
	end
	frame.masterIcon:SetWidth(10)
	frame.masterIcon:SetHeight(10)
end

local function getDisable(group)
	return PitBull_MasterLooter.db.profile[group].disable
end
local function setDisable(group, value)
	PitBull_MasterLooter.db.profile[group].disable = value
	
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		PitBull_MasterLooter:Update(unit, frame)
	end
end
PitBull_MasterLooter:RegisterPitBullOptionsMethod(function(group)
	if group == "player" or group == "party" or group == "raid" then
		return {
			type = 'group',
			name = 'Master looter',
			desc = 'Master looter',
			args = {
				hidden = {
					type = 'toggle',
					name = "Hide master looter icon",
					desc = "Hide master looter icon",
					get = getDisable,
					set = setDisable,
					passValue = group,
				}
			}
		}
	end
end)