local VERSION = tonumber(("$Revision: 32742 $"):match("%d+"))

local SharedMedia = AceLibrary("SharedMedia-1.0")

SharedMedia:Register("font", "ABF", "Interface\\AddOns\\PitBull\\Fonts\\ABF.ttf")

PitBull = AceLibrary("AceAddon-2.0"):new("AceDB-2.0", "AceEvent-2.0", "AceConsole-2.0", "AceModuleCore-2.0", "FuBarPlugin-2.0")
local PitBull, self = PitBull, PitBull
PitBull.version = "r" .. VERSION
PitBull.revision = VERSION
PitBull.date = ("$Date: 2007-04-15 17:01:52 -0400 (Sun, 15 Apr 2007) $"):match("%d%d%d%d%-%d%d%-%d%d")

local L = setmetatable({}, {__index = function(self, key) self[key] = key; return key end })
-- this is used as a placeholder for any localization I choose to do later.

local configMode = false
PitBull.configMode = configMode

local table_remove = table.remove
local select = select

local Dewdrop = AceLibrary("Dewdrop-2.0")
local Tablet = AceLibrary("Tablet-2.0")
local DogTag = AceLibrary("DogTag-1.0")

local new, del, newHash
do
	local list = setmetatable({}, {__mode='k'})
	function new(...)
		local t = next(list)
		if t then
			list[t] = nil
			for i = 1, select('#', ...) do
				t[i] = select(i, ...)
			end
			return t
		else
			return { ... }
		end
	end
	
	function newHash(...)
		local t = next(list)
		if t then
			list[t] = nil
		else
			t = {}
		end
		for i = 1, select('#', ...), 2 do
			t[select(i, ...)] = select(i+1, ...)
		end
		return t
	end
	
	function del(t)
		for k in pairs(t) do
			t[k] = nil
		end
		list[t] = true
		return nil
	end
end

local newFrame, delFrame
do
	local frameCache = {}
	local frameCount = {}
	function newFrame(kind, parent, extra)
		local frame
		if frameCache[kind] then
			frame = table_remove(frameCache[kind])
			if #frameCache[kind] == 0 then
				frameCache[kind] = del(frameCache[kind])
			end
			frame:SetParent(parent)
			if kind == "Texture" or kind == "FontString" then
				frame:SetDrawLayer(extra)
			end
			frame:Show()
		else	
			frameCount[kind] = (frameCount[kind] or 0) + 1
			local name = "PitBull" .. kind .. frameCount[kind]
			if kind == "Texture" then
				frame = parent:CreateTexture(name, extra)
			elseif kind == "FontString" then
				frame = parent:CreateFontString(name, extra)
			else
				frame = CreateFrame(kind, name, parent)
			end
		end
		return frame
	end
	
	function delFrame(frame)
		if type(frame) ~= "table" then
			error(("Bad argument #1 to `delFrame'. %s expected, got %s"):format("table", type(frame)), 2)
		end
		local kind = frame:GetObjectType()
		if kind == "FontString" then
			DogTag:UnregisterFontString(frame)
			frame:SetText("")
			frame:SetFontObject(nil)
			frame:SetNonSpaceWrap(true)
			frame:SetJustifyH("LEFT")
			frame:SetTextColor(1, 1, 1, 1)
		elseif kind == "Texture" then
			frame:SetTexture(nil)
			frame:SetVertexColor(1, 1, 1, 1)
			frame:SetBlendMode("BLEND")
			frame:SetDesaturated(false)
			frame:SetTexCoord(0, 1, 0, 1)
			frame:SetTexCoordModifiesRect(false)
		elseif kind == "StatusBar" then
			frame:SetStatusBarColor(1, 1, 1, 1)
			frame:SetStatusBarTexture(nil)
			frame:SetValue(1)
			frame:SetOrientation("HORIZONTAL")
		end
		frame:Hide()
		frame:SetParent(UIParent)
		frame:ClearAllPoints()
		frame:SetAlpha(1)
		local frameCache_kind = frameCache[kind]
		if not frameCache_kind then
			frameCache_kind = new()
			frameCache[kind] = frameCache_kind
		end
		frameCache_kind[#frameCache_kind+1] = frame
		return nil
	end
end

PitBull.new = new
PitBull.newHash = newHash
PitBull.del = del
PitBull.newFrame = newFrame
PitBull.delFrame = delFrame

local UnitIsUnit = UnitIsUnit
local UnitPowerType = UnitPowerType
local table_remove = table.remove

local frames = {}
local framesByUnit = {}
local framesByGroup = {}
local nonWackyFrames = {}
local allFrames = {}
local allFramesByUnit = {}
local allFramesByGroup = {}
local allFramesByCluster = {}
local groupMenus = {}
local wackyFrames = {}
local layouts, layoutNames = {}, {}
local borders, borderNames = {}, {}
local frameToModule = {}
local moduleOptionsMethods = {}
local clusters = {}

local function isframe(frame)
	return type(frame) == "table" and type(rawget(frame, 0)) == "userdata" and type(frame.IsObjectType) == "function"
end

local ProfilePitBull

PitBull:RegisterDB("PitBullDB")
do
	local colors = {
		rage = { 226/255, 45/255, 75/255 },
		energy = { 1, 220/255, 25/255 },
		focus = { 1, 210/255, 0 },
		mana = { 48/255, 113/255, 191/255 },
		
		unknown = { 0.8, 0.8, 0.8 },
		
		hostile = { 226/255, 45/255, 75/255 },
		neutral = { 1, 1, 34/255 },
		friendly = { 0.2, 0.8, 0.15 },
		civilian = { 48/255, 113/255, 191/255 },
		
		dead = { 0.6, 0.6, 0.6 },
		disconnected = { 0.7, 0.7, 0.7 },
		inCombat = { 1, 0, 0 },
		resting = { 1, 1, 0 },
		tapped = { 0.5, 0.5, 0.5 },
		
		petHappy = { 0, 1, 0 },
		petNeutral = { 1, 1, 0 },
		petAngry = { 1, 0, 0 },
		
		frameBG = { 0, 0, 0, 0.5 },
		frameBorder = { 1, 0.7, 1, 1 },
	}
	
	for k, v in pairs(RAID_CLASS_COLORS) do
		colors[k] = { v.r, v.g, v.b }
	end
	
	PitBull:RegisterDefaults('profile', {
		settings = {
			texture = "Blizzard",
			font = "ABF",
			fontsize = 12,
			locked = false,
			clamped = true,
			colors = colors,
			hideTooltipInCombat = false,
		},
		groups = {
			['**'] = {
				scale = 1,
				width = 200,
				height = 60,
				layout = "ABF",
				border = "Hidden",
				hidden = false,
				direction = "down",
				spacing = 10,
				freeform = false,
			},
			targettarget = {
				height = 20,
			},
			targettargettarget = {
				height = 20,
				hidden = true,
			},
			party = {
				nameSort = false,
				hidePartyInRaid = true,
			},
			pettarget = {
				hidden = true,
			},
			partytarget = {
				hidden = true,
			},
			partypettarget = {
				hidden = true,
			},
			raid = {
				groupStyle = "group",
				nameSort = true,
				show5manAsParty = true,
			},
			raidtarget = {
				hidden = true,
			},
			raidpet = {
				hidden = true,
			},
			mouseover = {
				hidden = true,
			},
			mouseovertarget = {
				hidden = true,
			},
		},
		units = {
			['*'] = {
				x = 0,
				y = 0,
			}
		},
		clusters = {
			['*'] = {
				x = 0,
				y = 0,
			}
		}
	})
end

local NumRaidMembers, NumPartyMembers = 0, 0

local function ShouldShowParty()
	if configMode == "raid" then
		return not PitBull.db.profile.groups.party.hidePartyInRaid
	end
	if NumRaidMembers > 0 then
		if NumRaidMembers == NumPartyMembers+1 and PitBull.db.profile.groups.raid.show5manAsParty then
			return true
		end
		return not PitBull.db.profile.groups.party.hidePartyInRaid
	end
	if configMode == "party" or NumPartyMembers > 0 then
		return true
	end
	return false
end

local function ShouldShowRaid()
	if configMode == "raid" then
		return true
	end
	if NumRaidMembers > 0 then
		if NumRaidMembers == NumPartyMembers+1 and PitBull.db.profile.groups.raid.show5manAsParty then
			return false
		end
		return true
	end
	return false
end

local function ShouldShowUnit(unit)
	if unit then
		if unit:find("^raid") then
			return ShouldShowRaid()
		elseif unit:find("^party") then
			return ShouldShowParty()
		end
		return true
	end
end

local IsLegitimateUnit = {player=true,pet=true,target=true,focus=true,mouseover=true}
for i = 1, 4 do
	IsLegitimateUnit['party' .. i] = true
	IsLegitimateUnit['partypet' .. i] = true
end
for i = 1, 40 do
	IsLegitimateUnit['raid' .. i] = true
	IsLegitimateUnit['raidpet' .. i] = true
end
setmetatable(IsLegitimateUnit, {__index=function(self, unit)
	if not unit:find("target$") then
		self[unit] = false
		return false
	end
	local nonTarget = unit:sub(1, -7)
	local good = self[nonTarget]
	self[unit] = good
	return good
end})
PitBull.IsLegitimateUnit = IsLegitimateUnit

local IsWackyUnit = setmetatable({player=false, target=false, pet=false, mouseover=false, focus=false}, {__index=function(self, unit)
	if not unit then
		AceLibrary("AceConsole-2.0"):Print("Please report this to ckknight:", debugstack())
	end
	local party = unit:match("^party(%d)$") or unit:match("^partypet(%d)$")
	if party then
		party = tonumber(party)
		if party and party >= 1 and party <= 4 then
			self[unit] = false
			return false
		end
		self[unit] = true
		return true
	end
	local raid = unit:match("^raid(%d%d?)$")
	if raid then
		raid = tonumber(raid)
		if raid and raid >= 1 and raid <= 40 then
			self[unit] = false
			return false
		end
		self[unit] = true
		return true
	end
	self[unit] = true
	return true
end})
PitBull.IsWackyUnit = IsWackyUnit
local IsWackyGroup = setmetatable({player=false, target=false, pet=false, mouseover=false, focus=false, party=false, partypet=false, raid=false,}, {__index=function(self, group)
	self[group] = true
	return true
end})
PitBull.IsWackyGroup = IsWackyGroup

local GroupToLocale = {player = L["Player"], target = L["Target"], pet = L["Player's pet"], party = L["Party"], partypet = L["Party pets"], raid = L["Raid"], raidpet = L["Raid pets"], mouseover = L["Mouse-over"], focus = L["Focus"]}
setmetatable(GroupToLocale, {__index=function(self, group)
	if not group:find("target$") then
		self[group] = group
		return group
	end
	local nonTarget = group:sub(1, -7)
	local good
	if group:find("^player") or group:find("^pet") or group:find("^mouseover") or group:find("^target") or group:find("^focus") then
		good = L["%s's target"]:format(self[nonTarget])
	else
		good = L["%s targets"]:format(self[nonTarget])
	end
	self[group] = good
	return good
end})
PitBull.GroupToLocale = GroupToLocale

-- FuBar configuration
PitBull.hasIcon = true
PitBull.hideWithoutStandby = true
PitBull.defaultPosition = "MINIMAP"
PitBull.defaultMinimapPosition = 200
PitBull.hasNoText = true
PitBull.hasNoColor = true
PitBull.cannotDetachTooltip = true
PitBull.independentProfile = true

function PitBull:OnInitialize()
	self.colorConstants = self.db.profile.settings.colors
	-- REMOVE LATER - added April 13, 2007
	local hidePartyInRaid = self.db.profile.settings.hidePartyInRaid
	if hidePartyInRaid ~= nil then
		self.db.profile.groups.party.hidePartyInRaid = hidePartyInRaid
	end
	-- REMOVE LATER
end

local ProfilePitBull_open = false
do
	local function colorSet(key, r, g, b, a)
		local t = self.db.profile.settings.colors[key]
		t[1] = r
		t[2] = g
		t[3] = b
		t[4] = a
		self:UpdateAll()
	end

	local function colorGet(key)
		return unpack(self.db.profile.settings.colors[key])
	end

	PitBull.options = {
		type = 'group',
		handler = PitBull,
		args = {
			new = {
				name = "New unit frame",
				desc = "Create a new unit frame",
				type = 'text',
				usage = "something like {focustarget || targettargettarget || etc.}",
				get = false,
				set = "CreateUnitFrame",
				validate = function(value)
					return IsLegitimateUnit[value]
				end,
				disabled = InCombatLockdown,
			},
			global = {
				name = "Global settings",
				desc = "Global settings",
				type = 'group',
				args = {
					texture = {
						name = "StatusBar texture",
						desc = "StatusBar texture",
						type = 'text',
						validate = SharedMedia:List('statusbar'),
						get = function()
							return self.db.profile.settings.texture
						end,
						set = function(value)
							self.db.profile.settings.texture = value
							self:UpdateAllStatusBarTextures()
						end,
					},
					font = {
						type = "group",
						name = "Font",
						desc = "Font Settings",
						args = {
							fonttype = {
								name = "Type",
								desc = "Type",
								type = 'text',
								validate = SharedMedia:List('font'),
								get = function()
									return self.db.profile.settings.font
								end,
								set = function(value)
									self.db.profile.settings.font = value
									self:UpdateAllFonts()
								end,
							},
							fontsize = {
								name = "Size",
								desc = "Size",
								type = "range",
								get = function()
									return self.db.profile.settings.fontsize
								end,
								set = function(value)
									self.db.profile.settings.fontsize = value
									self:UpdateAllFonts()
								end,
								min = 8, max = 32, step = 1,
							},
						},
					},			
					locked = {
						name = "Locked",
						desc = "Make it so the frames cannot be moved",
						type = 'toggle',
						get = function()
							return self.db.profile.settings.locked
						end,
						set = function(value)
							self.db.profile.settings.locked = value
						end,
					},
					clamped = {
						name = "Clamp frames to screen",
						desc = "Make it so that frames cannot be dragged off-screen",
						type = 'toggle',
						get = "AreUnitFramesClampedToScreen",
						set = "ToggleUnitFramesClampedToScreen",
						handler = self,
						disabled = InCombatLockdown,
					},
					configMode = {
						name = "Configuration mode",
						desc = "Show all frames that can be shown, for easy configuration.",
						type = 'text',
						validate = {
							solo = "Solo",
							party = "Party",
							raid = "Raid",
							disabled = "Disabled"
						},
						get = function()
							return configMode or "disabled"
						end,
						set = "ChangeConfigMode",
						handler = self,
						disabled = InCombatLockdown,
					},
					hideTooltipInCombat = {
						name = "Hide tooltips in combat",
						desc = "Hide tooltips in combat",
						type = 'toggle',
						get = function()
							return self.db.profile.settings.hideTooltipInCombat
						end,
						set = function(value)
							PitBull.db.profile.settings.hideTooltipInCombat = value
						end,
					},
					colors = {
						name = "Colors",
						desc = "Colors",
						type = 'group',
						args = {
							power = {
								name = "Power",
								desc = "Power",
								type = 'group',
								pass = true,
								get = colorGet,
								set = colorSet,
								args = {
									rage = {
										name = "Rage",
										desc = "Rage",
										type = 'color'
									},
									mana = {
										name = "Mana",
										desc = "Mana",
										type = 'color'
									},
									focus = {
										name = "Focus",
										desc = "Focus",
										type = 'color'
									},
									energy = {
										name = "Energy",
										desc = "Energy",
										type = 'color'
									},
								}
							},
							class = {
								name = "Classes",
								desc = "Classes",
								type = 'group',
								pass = true,
								get = colorGet,
								set = colorSet,
								args = {
									WARLOCK = {
										name = "Warlock",
										desc = "Warlock",
										type = 'color'
									},
									PRIEST = {
										name = "Priest",
										desc = "Priest",
										type = 'color'
									},
									WARRIOR = {
										name = "Warrior",
										desc = "Warrior",
										type = 'color'
									},
									PALADIN = {
										name = "Paladin",
										desc = "Paladin",
										type = 'color'
									},
									SHAMAN = {
										name = "Shaman",
										desc = "Shaman",
										type = 'color'
									},
									MAGE = {
										name = "Mage",
										desc = "Mage",
										type = 'color'
									},
									DRUID = {
										name = "Druid",
										desc = "Druid",
										type = 'color'
									},
									ROGUE = {
										name = "Rogue",
										desc = "Rogue",
										type = 'color'
									},
									HUNTER = {
										name = "Hunter",
										desc = "Hunter",
										type = 'color'
									},
								}
							},
							reaction = {
								name = "Reaction",
								desc = "Reaction",
								type = 'group',
								pass = true,
								get = colorGet,
								set = colorSet,
								args = {
									hostile = {
										name = "Hostile",
										desc = "Hostile",
										type = 'color',
									},
									neutral = {
										name = "Neutral",
										desc = "Neutral",
										type = 'color',
									},
									friendly = {
										name = "Friendly",
										desc = "Friendly",
										type = 'color',
									},
									civilian = {
										name = "Civilian",
										desc = "Civilian",
										type = 'color',
									},
								}
							},
							petHappiness = {
								name = "Pet happiness",
								desc = "Pet happiness",
								type = 'group',
								pass = true,
								get = colorGet,
								set = colorSet,
								args = {
									petHappy = {
										name = "Happy",
										desc = "Happy",
										type = 'color',
									},
									petNeutral = {
										name = "Neutral",
										desc = "Neutral",
										type = 'color',
									},
									petAngry = {
										name = "Angry",
										desc = "Angry",
										type = 'color',
									},
								}
							},
							other = {
								name = "Other",
								desc = "Other",
								type = 'group',
								pass = true,
								get = colorGet,
								set = colorSet,
								args = {
									dead = {
										name = "Dead or Ghost",
										desc = "Dead or Ghost",
										type = 'color',
									},
									disconnected = {
										name = "Disconnected",
										desc = "Disconnected",
										type = 'color',
									},
									inCombat = {
										name = "In combat",
										desc = "In combat",
										type = 'color',
									},
									resting = {
										name = "Resting",
										desc = "Resting",
										type = 'color',
									},
									tapped = {
										name = "Tapped",
										desc = "Tapped",
										type = 'color',
									},
									unknown = {
										name = "Unknown",
										desc = "Unknown",
										type = 'color'
									},
								}
							},
							frame = {
								name = "Frame colors",
								desc = "Frame colors",
								type = 'group',
								args = {
									background = {
										name = "Background",
										desc = "Background",
										type = 'color',
										get = function()
											return unpack(self.db.profile.settings.colors.frameBG)
										end,
										set = function(r, g, b, a)
											local t = self.db.profile.settings.colors.frameBG
											t[1] = r
											t[2] = g
											t[3] = b
											t[4] = a
											self:UpdateAllBackdropsAndBorders()
										end,
										hasAlpha = true,
									},
									border = {
										name = "Border",
										desc = "Border",
										type = 'color',
										get = function()
											return unpack(self.db.profile.settings.colors.frameBorder)
										end,
										set = function(r, g, b, a)
											local t = self.db.profile.settings.colors.frameBorder
											t[1] = r
											t[2] = g
											t[3] = b
											t[4] = a
											self:UpdateAllBackdropsAndBorders()
										end,
										hasAlpha = true,
									},
								}
							}
						}
					},
				}
			},
			debug = {
				order = -1,
				type = 'group',
				name = "Debugging",
				desc = "Debugging capabilities",
				args = {
					toggle = {
						order = 50,
						name = "Enable profiling",
						desc = "Enable profiling of PitBull. (Note: changing causes UI to reload)",
						type = 'toggle',
						get = function()
							return PitBull.db.account.debug
						end,
						set = function(value)
							PitBull.db.account.debug = value
							ReloadUI()
						end
					},
					profile = {
						name = "Display profiling results",
						desc = "Show the memory and times for each function of PitBull",
						type = 'toggle',
						get = function()
							return ProfilePitBull_open
						end,
						set = function()
							ProfilePitBull()
						end,
						hidden = function() return not ProfilePitBull end,
					}
				},
			},
		},
	}
end
PitBull.OnMenuRequest = PitBull.options

local modulesWithFunction = setmetatable({}, {__index = function(self, key)
	local t = {}
	for name, module in PitBull:IterateModules() do
		if type(module[key]) == "function" then
			t[#t+1] = module
		end
	end
	self[key] = t
	
	return t
end})
local SetupDebugging
function PitBull:OnEnable(first)
	if not ClickCastFrames then
		ClickCastFrames = {}
	end
	
	if PitBull.db.account.debug then
		SetupDebugging()
	end
		
	Dewdrop:InjectAceOptionsTable(self, self.options)
	
	self:AddGroupToAceOptions("player")
	if not self.db.profile.groups.player.hidden then
		self:CreateUnitFrame("player")
	end
	self:AddGroupToAceOptions("pet")
	if not self.db.profile.groups.pet.hidden then
		self:CreateUnitFrame("pet")
	end
	self:AddGroupToAceOptions("pettarget")
	if not self.db.profile.groups.pettarget.hidden then
		self:CreateUnitFrame("pettarget")
	end
	self:AddGroupToAceOptions("target")
	if not self.db.profile.groups.target.hidden then
		self:CreateUnitFrame("target")
	end
	self:AddGroupToAceOptions("targettarget")
	if not self.db.profile.groups.targettarget.hidden then
		self:CreateUnitFrame("targettarget")
	end
	self:AddGroupToAceOptions("targettargettarget")
	if not self.db.profile.groups.targettargettarget.hidden then
		self:CreateUnitFrame("targettargettarget")
	end
	self:AddGroupToAceOptions("focus")
	if not self.db.profile.groups.focus.hidden then
		self:CreateUnitFrame("focus")
	end
	self:AddGroupToAceOptions("focustarget")
	if not self.db.profile.groups.focustarget.hidden then
		self:CreateUnitFrame("focustarget")
	end
	self:AddGroupToAceOptions("party")
	self:AddGroupToAceOptions("partypet")
	self:AddGroupToAceOptions("partytarget")
	self:AddGroupToAceOptions("partypettarget")
	self:AddGroupToAceOptions("raid")
	self:AddGroupToAceOptions("raidpet")
	self:AddGroupToAceOptions("raidtarget")
	self:AddGroupToAceOptions("mouseover")
	if not self.db.profile.groups.mouseover.hidden then
		self:CreateUnitFrame("mouseover")
	end
	self:AddGroupToAceOptions("mouseovertarget")
	if not self.db.profile.groups.mouseovertarget.hidden then
		self:CreateUnitFrame("mouseovertarget")
	end
	
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:RegisterEvent("PLAYER_FOCUS_CHANGED")
	self:RegisterEvent("UPDATE_MOUSEOVER_UNIT")
	self:RegisterEvent("AceEvent_FullyInitialized", function()
		self:RegisterEvent("PARTY_MEMBERS_CHANGED")
		self:RegisterEvent("RAID_ROSTER_UPDATE", "PARTY_MEMBERS_CHANGED")
		self:RegisterEvent("UNIT_PET", "PARTY_MEMBERS_CHANGED")
		self:PARTY_MEMBERS_CHANGED()
		self:UpdateAll()
	end)
	self:RegisterEvent("PLAYER_PET_CHANGED")
	self:RegisterEvent("SharedMedia_Registered")
	self:RegisterEvent("PLAYER_REGEN_DISABLED")
	self:RegisterEvent("UNIT_HAPPINESS")
	
	self:UpdateAll()
	
	self:ScheduleRepeatingEvent("PitBull-UpdateWackyFrames", self.UpdateWackyFrames, 0.15, self)
	
	self:RegisterChatCommand('/PitBull', '/Pit', {
		type = 'execute',
		func = function()
			Dewdrop:Open("PitBull")
		end
	})
	
	Dewdrop:Register("PitBull",
		'children', PitBull.options,
		'cursorX', true,
		'cursorY', true
	)
end

local function getScale(group)
	return self.db.profile.groups[group].scale
end
local function getWidth(group)
	return self.db.profile.groups[group].width
end
local function getHeight(group)
	return self.db.profile.groups[group].height
end
local function getLayout(group)
	return self.db.profile.groups[group].layout
end
local function getBorder(group)
	return self.db.profile.groups[group].border
end
local function getHidden(group)
	return self.db.profile.groups[group].hidden
end
local function getDirection(group)
	return self.db.profile.groups[group].direction
end
local function getSpacing(group)
	return self.db.profile.groups[group].spacing
end
local function getFreeform(group)
	return self.db.profile.groups[group].freeform
end
function PitBull:AddGroupToAceOptions(group)
	if self.options.args[group] then
		return
	end
	groupMenus[group] = true
	local order
	if group:find("^player") then
		order = 50 + group:len()/1000
	elseif group:find("^pet") then
		order = 51 + group:len()/1000
	elseif group:find("^target") then
		order = 52 + group:len()/1000
	elseif group:find("^partypet") then
		order = 54 + group:len()/1000
	elseif group:find("^party") then
		order = 53 + group:len()/1000
	elseif group:find("^raidpet") then
		order = 56 + group:len()/1000
	elseif group:find("^raid") then
		order = 55 + group:len()/1000
	elseif group:find("^focus") then
		order = 57 + group:len()/1000
	elseif group:find("^mouseover") then
		order = 58 + group:len()/1000
	else
		order = 59 + group:len()/1000
	end
	self.options.args[group] = {
		order = order,
		name = GroupToLocale[group],
		desc = GroupToLocale[group],
		type = 'group',
		args = {
			header = {
				order = 1,
				name = GroupToLocale[group],
				type = 'header',
			},
			size = {
				name = "Size",
				desc = "Size of the unit frames",
				type = 'group',
				args = {
					scale = {
						name = "Scale",
						desc = "Scale of the unit frame",
						type = 'range',
						isPercent = true,
						min = 0.5,
						max = 2,
						step = 0.05,
						get = getScale,
						set = "ChangeScale",
						passValue = group,
						handler = PitBull,
						disabled = InCombatLockdown,
					},
					width = {
						name = "Width",
						desc = "Width of the unit frame",
						type = 'range',
						min = 20,
						max = 400,
						step = 5,
						get = getWidth,
						set = "ChangeWidth",
						passValue = group,
						handler = PitBull,
						disabled = InCombatLockdown,
					},
					height = {
						name = "Height",
						desc = "Height of the unit frame",
						type = 'range',
						min = 10,
						max = 400,
						step = 5,
						get = getHeight,
						set = "ChangeHeight",
						passValue = group,
						handler = PitBull,
						disabled = InCombatLockdown,
					},
				},
				disabled = InCombatLockdown,
			},
			layout = {
				name = "Layout",
				desc = "Layout",
				type = 'text',
				validate = layoutNames,
				get = getLayout,
				set = "ChangeLayout",
				passValue = group,
				handler = PitBull,
			},
			layoutSettings = {
				name = "Layout settings",
				desc = "Layout settings",
				type = 'group',
				args = {},
				hidden = function()
					if not next(self.options.args[group].args.layoutSettings.args) then
						return true
					end
					for k,v in pairs(self.options.args[group].args.layoutSettings.args) do
						if not v.hidden then
							return false
						elseif type(v.hidden) == "function" then
							local val = v.hidden(v.passValue)
							if not val then
								return false
							end
						elseif v.hidden ~= true then
							return false
						end
					end
					return true
				end
			},
			border = {
				name = "Border",
				desc = "Border",
				type = 'text',
				validate = borderNames,
				get = getBorder,
				set = "ChangeBorder",
				passValue = group,
				handler = PitBull,
			},
			hidden = {
				name = "Hidden",
				desc = "Hidden",
				type = 'toggle',
				get = getHidden,
				set = "ChangeHidden",
				passValue = group,
				handler = PitBull,
				disabled = InCombatLockdown
			}
		},
	}
	if group == "party" then
		self.options.args[group].args.hidePartyInRaid = {
			name = "Hide party frames in raid",
			desc = "Hide party frames in raid",
			type = 'toggle',
			get = function()
				return self.db.profile.groups.party.hidePartyInRaid
			end,
			set = "ChangeHidePartyInRaid",
			handler = self,
			disabled = InCombatLockdown,
		}
	elseif group == "raid" then
		self.options.args[group].args.show5manAsParty = {
			name = "Show 5-man raid as a party.",
			desc = "Show a 5-man, 1-party raid as a party instead of a raid. This is useful for the Arena.",
			type = 'toggle',
			get = function()
				return self.db.profile.groups.raid.show5manAsParty
			end,
			set = "ChangeShow5manAsParty",
			handler = self,
			disabled = InCombatLockdown,
		}
	end
	if group:find("^party") or group:find("^raid") then
		self.options.args[group].args.grouping = {
			name = "Grouping",
			desc = "Grouping",
			type = 'group',
			args = {
--[[				freeform = {
					name = "Enable freeform movement",
					desc = "Enable freeform movement",
					type = 'toggle',
					get = getFreeform,
					set = "ChangeGroupFreeform",
					passValue = group,
					handler = PitBull,
					disabled = InCombatLockdown,
				},]]
				direction = {
					name = "Direction",
					desc = "Direction",
					type = 'text',
					validate = {
						left = "Left",
						right = "Right",
						up = "Up",
						down = "Down",
					},
					get = getDirection,
					set = "ChangeGroupDirection",
					passValue = group,
					handler = PitBull,
					disabled = InCombatLockdown
				},
				spacing = {
					name = "Spacing",
					desc = "Spacing",
					type = 'range',
					min = 0,
					max = 200,
					step = 5,
					get = getSpacing,
					set = "ChangeGroupSpacing",
					passValue = group,
					handler = PitBull,
					disabled = InCombatLockdown
				},
			}
		}
		if group == "party" or group == "raid" then
			self.options.args[group].args.grouping.args.nameSort = {
				name = "Sort by name",
				desc = "Whether to sort by name or by index",
				type = 'toggle',
				get = function(group)
					return self.db.profile.groups[group].nameSort
				end,
				set = "ChangeNameSort",
				passValue = group,
				disabled = InCombatLockdown,
				handler = PitBull,
			}
		end
		if group == "raid" then
			self.options.args[group].args.grouping.args.groupStyle = {
				name = "Style",
				desc = "Style to group by",
				type = 'text',
				validate = {
					class = "Class",
					group = "Raid Group",
				},
				get = function()
					return self.db.profile.groups.raid.groupStyle
				end,
				set = "ChangeRaidGroupStyle",
				disabled = InCombatLockdown,
				handler = PitBull,
			}
		end
	end
	if not self.options.args.sep then
		self.options.args.sep = {
			order = 60,
			type = 'header'
		}
	end
	
	for module, method in pairs(moduleOptionsMethods) do
		local name = module.name
	
		local options = PitBull.options.args[group].args
		if type(method) == "function" then
			options[name] = method(group)
		else
			options[name] = module[method](module, group)
		end
		if options[name] and not options[name].handler then
			options[name].handler = module
		end
	end
end

function PitBull:OnModuleCreated(name, module)
	for k in pairs(modulesWithFunction) do
		modulesWithFunction[k] = nil
	end
end

function PitBull:OnTooltipUpdate()
	local cat = Tablet:AddCategory(
		'columns', 2,
		'child_textR', 1,
		'child_textG', 1,
		'child_textB', 0,
		'child_text2R', 1,
		'child_text2G', 1,
		'child_text2B', 1
	)
	cat:AddLine(
		'text', LOCKED .. ":",
		'text2', self.db.profile.settings.locked and YES or NO
	)
	local text
	if configMode == "solo" then
		text = "Solo"
	elseif configMode == "party" then
		text = "Party"
	elseif configMode == "raid" then
		text = "Raid"
	else
		text = "Off"
	end
	cat:AddLine(
		'text', "Configuration mode:",
		'text2', text
	)
	Tablet:SetHint("|cffffff00Click|r to lock/unlock unit frames" .. "\n" .. "|cffffff00Shift-click|r to change configuration mode" .. "\n" .. "|cffffff00Right-click|r to change settings")
end

function PitBull:OnClick()
	local shift, alt, ctrl = IsShiftKeyDown(), IsAltKeyDown(), IsControlKeyDown()
	if not shift and not alt and not ctrl then
		self.db.profile.settings.locked = not self.db.profile.settings.locked
		self:UpdateTooltip()
	elseif shift and not alt and not ctrl then
		if NumPartyMembers > 0 or NumRaidMembers > 0 then
			if configMode then
				self:ChangeConfigMode(nil)
			else
				self:ChangeConfigMode(true)
			end
		else
			if not configMode then
				self:ChangeConfigMode("solo")
			elseif configMode == "solo" then
				self:ChangeConfigMode("party")
			elseif configMode == "party" then
				self:ChangeConfigMode("raid")
			else
				self:ChangeConfigMode(nil)
			end
		end
		self:UpdateTooltip()
	end
end

function PitBull:ReorganizeCluster(cluster)
	local header = clusters[cluster]
	local direction = self.db.profile.groups[header.group].direction
	local spacing = self.db.profile.groups[header.group].spacing
	local point
	if direction == "up" then
		header:SetAttribute("xOffset", 0)
		header:SetAttribute("yOffset", spacing)
		point = "BOTTOM"
	elseif direction == "left" then
		header:SetAttribute("xOffset", -spacing)
		header:SetAttribute("yOffset", 0)
		point = "RIGHT"
	elseif direction == "right" then
		header:SetAttribute("xOffset", spacing)
		header:SetAttribute("yOffset", 0)
		point = "LEFT"
	else
		header:SetAttribute("xOffset", 0)
		header:SetAttribute("yOffset", -spacing)
		point = "TOP"
	end
	header:SetAttribute("point", point)
	local uis = UIParent:GetScale()
	local s = header:GetEffectiveScale()
	local db = self.db.profile.clusters[cluster]
	header:ClearAllPoints()
	header:SetPoint(point, UIParent, "CENTER", db.x*uis/s, db.y*uis/s)
end

function PitBull:PLAYER_REGEN_DISABLED()
	if configMode then
		self:Print("Turning off config mode due to entering combat. Feel free to re-enable it once you are out of combat.")
		
		self:ChangeConfigMode(nil)
	end
end

function PitBull:PLAYER_TARGET_CHANGED()
	if allFramesByUnit.target and (configMode or UnitExists('target')) then
		self:PopulateUnitFrame(allFramesByUnit.target)
	elseif framesByUnit.target then
		self:ClearUnitFrame(framesByUnit.target)
	end
end

function PitBull:PLAYER_FOCUS_CHANGED()
	if allFramesByUnit.focus and (configMode or UnitExists('focus')) then
		self:PopulateUnitFrame(allFramesByUnit.focus)
	elseif framesByUnit.focus then
		self:ClearUnitFrame(framesByUnit.focus)
	end
end

function PitBull:UPDATE_MOUSEOVER_UNIT()
	if allFramesByUnit.mouseover and (configMode or UnitExists('mouseover')) then
		self:PopulateUnitFrame(allFramesByUnit.mouseover)
	elseif framesByUnit.mouseover then
		self:ClearUnitFrame(framesByUnit.mouseover)
	end
end

local lastRaid, lastParty = 0, 0
function PitBull:PARTY_MEMBERS_CHANGED() -- also RAID_ROSTER_UPDATE and UNIT_PET
	NumRaidMembers = GetNumRaidMembers()
	NumPartyMembers = GetNumPartyMembers()
	if configMode and event then
		if configMode == "raid" then
			if NumRaidMembers == 0 then
				self:Print("Turning off Configuration mode, due to real group change. Feel free to re-enable.")
				self:ChangeConfigMode(nil)
			end
		elseif configMode == "party" then
			if NumRaidMember > 0 or NumPartyMembers == 0 then
				self:Print("Turning off Configuration mode, due to real group change. Feel free to re-enable.")
				self:ChangeConfigMode(nil)
			end
		elseif configMode == "solo" then
			if NumRaidMember > 0 or NumPartyMembers > 0 then
				self:Print("Turning off Configuration mode, due to real group change. Feel free to re-enable.")
				self:ChangeConfigMode(nil)
			end
		end
	end
	
	local showparty, showraid
	if ShouldShowParty() then
		showparty = true
		for group, db in pairs(self.db.profile.groups) do
			if group:find("^party") and not db.hidden then
				self:ScheduleLeaveCombatAction(self, 'CreateUnitCluster', group)
			end
		end
	else
		for cluster in pairs(clusters) do
			if cluster:find("^party") then
				self:ScheduleLeaveCombatAction(self, 'DestroyUnitCluster', cluster)
			end
		end
	end
	
	if ShouldShowRaid() then
		showraid = true
		for group, db in pairs(self.db.profile.groups) do
			if group:find("^raid") and not db.hidden then
				self:ScheduleLeaveCombatAction(self, 'CreateUnitCluster', group)
			end
		end
	else
		for cluster in pairs(clusters) do
			if cluster:find("^raid") then
				self:ScheduleLeaveCombatAction(self, 'DestroyUnitCluster', cluster)
			end
		end
	end
	
	
	for frame in pairs(allFrames) do
		local unit = frame:GetUnit()
		if not unit or self.db.profile.groups[frame.group].hidden then
			self:ClearUnitFrame(frame)
		elseif showparty and unit:find("^party") then
			if configMode or UnitExists(unit) then
				self:PopulateUnitFrame(frame)
			else
				self:ClearUnitFrame(frame)
			end
		elseif showraid and unit:find("^raid") then
			if configMode or UnitExists(unit) then
				self:PopulateUnitFrame(frame)
			else
				self:ClearUnitFrame(frame)
			end
		end
	end
	
	if not self.db.profile.groups['pet'].hidden and (configMode or UnitExists('pet')) then
		self:PopulateUnitFrame(allFramesByUnit.pet)
	elseif framesByUnit.pet then
		self:ClearUnitFrame(framesByUnit.pet)
	end
	
	lastRaid, lastParty = raid, party
end

function PitBull:PLAYER_PET_CHANGED()
	if not self.db.profile.groups['pet'].hidden and (configMode or UnitExists('pet')) then
		self:PopulateUnitFrame(allFramesByUnit.pet)
	elseif framesByUnit.pet then
		self:ClearUnitFrame(framesByUnit.pet)
	end
end

function PitBull:UNIT_HAPPINESS()
	if framesByUnit.pet then
		self:UpdateFrame(framesByUnit.pet)
	end
end

function PitBull:UpdateAll()
	for frame in pairs(frames) do
		local unit = frame:GetUnit()
		if (configMode or UnitExists(unit)) and ShouldShowUnit(unit) and not self.db.profile.groups[frame.group].hidden then
			self:UpdateLayout(frame)
			self:UpdateBackdropAndBorder(frame)
			self:UpdateFrame(frame)
		end
	end
	self:UpdateAllStatusBarTextures()
	self:UpdateAllFonts()
end

function PitBull:UpdateWackyFrames()
	-- this is to update frames that are not handled by Blizzard's standard events, e.g. targettarget, targettargettarget
	for frame in pairs(wackyFrames) do
		local unit = frame:GetUnit()
		if configMode or UnitExists(unit) then
			self:UpdateFrame(frame)
		end
	end
end

function PitBull:UpdateFrame(frame)
	-- update everything about a frame
	if not isframe(frame) then
		error(("Bad argument #2 to `UpdateFrame'. Expected %s, got %s"):format("frame", type(frame)), 2)
	end
	if not frames[frame] then
		return
	end
	local unit = frame:GetUnit()
	
	for _, module in ipairs(modulesWithFunction.OnUpdateFrame) do
		module:OnUpdateFrame(unit, frame)
	end
	
	DogTag:UpdateAllForUnit(unit)
end

local ignore = { overlay = true, db = true, group = true, unit = true, backdrop = true, border1 = true, border2 = true, border3 = true, border4 = true, border5 = true, border6 = true, border7 = true, border8 = true }
local layoutsToUpdate
function PitBull:_UpdateLayout(frame)
	for frame in pairs(layoutsToUpdate) do
		layoutsToUpdate[frame] = nil
		if frames[frame] then
			local unit = frame:GetUnit()
		
			if not frame.group then
				AceLibrary("AceConsole-2.0"):Print(frame)
			end
			local layoutName = self.db.profile.groups[frame.group].layout
			local current_layout = layouts[layoutName]
			if not current_layout then
				layoutName = "ABF"
				current_layout = layouts.ABF
			end

			if current_layout.extraFrames then
				for _, v in ipairs(current_layout.extraFrames) do
					if frame[v] then
						frame[v] = delFrame(frame[v])
					end
				end
			end

			local tmp, tmp2 = new(), new()

			for k, v in pairs(frame) do
				if not ignore[k] and type(k) == "string" and isframe(v) then
					tmp2[k] = v
					tmp[k] = v:IsShown()
					v:Show()
					v:ClearAllPoints()
					v:SetWidth(0)
					v:SetHeight(0)
					if v.SetScale then
						v:SetScale(1)
					end
					local objectType = v:GetObjectType()
					if objectType == "FontString" then
						local fontObject = v:GetFontObject()
						if fontObject then
							v:SetFont(fontObject:GetFont())
						end
						v:SetJustifyH("LEFT")
						v:SetHeight(1)
					elseif objectType == "StatusBar" then
						v:SetOrientation("HORIZONTAL")
					end
				end
			end
			for k, v in pairs(tmp2) do
				tmp2[k] = nil
				local func = current_layout[k]
				if type(func) == "function" then
					local success, ret = pcall(func, unit, frame)
					if not success then
						self:Print("There was an error with the layout %q, preventing the frame for %q to be shown properly. Please inform the author: %s", layoutName, unit, ret)
						v:Hide()
					end
				else
					local module = frameToModule[k]
					if module and type(module.OnUnknownLayout) == "function" then
						module:OnUnknownLayout(unit, frame, k)
					end
				end
				if tmp[k] ~= v:IsShown() then
					-- change of visibility
					if v:IsObjectType("FontString") and v:IsShown() then
						DogTag:UpdateFontString(v)
					end
				end
				tmp[k] = nil
			end
			tmp, tmp2 = del(tmp), del(tmp2)
			if type(current_layout.always) == "function" then
				local success, ret = pcall(current_layout.always, unit, frame)
				if not success then
					self:Print("There was an error with the layout %q, preventing the function \"always\" for %q from being called properly. Please inform the author: %s", layoutName, unit, ret)
				end
			end
			for _, module in ipairs(modulesWithFunction.OnUpdateLayout) do
				module:OnUpdateLayout(unit, frame)
			end

			self:UpdateLayoutSettings(frame.group)
		end
	end
	layoutsToUpdate = del(layoutsToUpdate)
end

function PitBull:UpdateLayout(frame)
	if not isframe(frame) then
		error(("Bad argument #2 to `UpdateLayout'. Expected %s, got %s"):format("frame", type(frame)), 2)
	end
	if not layoutsToUpdate then
		layoutsToUpdate = new()
		self:ScheduleEvent("PitBull-UpdateLayout", self._UpdateLayout, 0, self)
	end
		
	layoutsToUpdate[frame] = true
end

function PitBull:UpdateAllLayouts()
	for unit, frame in self:IterateUnitFrames() do
		self:UpdateLayout(frame)
	end
end

function PitBull:UpdateBackdropAndBorder(frame)
	if not isframe(frame) then
		error(("Bad argument #2 to `UpdateBackdropAndBorder'. Expected %s, got %s"):format("frame", type(frame)), 2)
	end
	local unit = frame:GetUnit()
	if frame.backdrop then
		frame.backdrop = delFrame(frame.backdrop)
	end
	if frame.border1 then
		for i = 1, 8 do
			frame['border' .. i] = delFrame(frame['border' .. i])
		end
	end
	local border = borders[self.db.profile.groups[frame.group].border] or borders.Hidden
	local edgeFile = border[1] or ""
	local edgeSize = border[2] or 0
	local inset = border[3] or 3
	local bg = newFrame("Texture", frame, "BACKGROUND")
	frame.backdrop = bg
	bg:SetTexture("Interface\\ChatFrame\\ChatFrameBackground")
	bg:SetVertexColor(unpack(self.db.profile.settings.colors.frameBG))
	bg:SetHeight(16)
	bg:SetWidth(16)
	bg:SetPoint("TOPLEFT", frame, "TOPLEFT", inset - 5, -inset + 5)
	bg:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT", -inset + 5, inset - 5)
	if edgeFile ~= "" and edgeSize ~= 0 then
		local border = newFrame("Texture", frame, "BORDER")
		frame.border1 = border
		border:SetTexture(edgeFile)
		border:SetVertexColor(unpack(self.db.profile.settings.colors.frameBorder))
		border:SetWidth(edgeSize)
		border:SetHeight(edgeSize)
		border:SetPoint("TOPLEFT", frame, "TOPLEFT", -5, -edgeSize + 5)
		border:SetPoint("BOTTOMLEFT", frame, "BOTTOMLEFT", -5, edgeSize - 5)
		border:SetTexCoord(0, 0, 0, 1.948, 0.125, 0, 0.125, 1.948)
		local border = newFrame("Texture", frame, "BORDER")
		frame.border2 = border
		border:SetTexture(edgeFile)
		border:SetVertexColor(unpack(self.db.profile.settings.colors.frameBorder))
		border:SetWidth(edgeSize)
		border:SetHeight(edgeSize)
		border:SetPoint("TOPRIGHT", frame, "TOPRIGHT", 5, -edgeSize + 5)
		border:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT", 5, edgeSize - 5)
		border:SetTexCoord(0.125, 0, 0.125, 1.948, 0.25, 0, 0.25, 1.948)
		local border = newFrame("Texture", frame, "BORDER")
		frame.border3 = border
		border:SetTexture(edgeFile)
		border:SetVertexColor(unpack(self.db.profile.settings.colors.frameBorder))
		border:SetWidth(edgeSize*8)
		border:SetHeight(edgeSize)
		border:SetPoint("TOPLEFT", frame, "TOPLEFT", edgeSize - 5, 5)
		border:SetPoint("TOPRIGHT", frame, "TOPRIGHT", -edgeSize + 5, 5)
		border:SetTexCoord(0.25, 9.2808, 0.375, 9.2808, 0.25, 0, 0.375, 0)
		local border = newFrame("Texture", frame, "BORDER")
		frame.border4 = border
		border:SetTexture(edgeFile)
		border:SetVertexColor(unpack(self.db.profile.settings.colors.frameBorder))
		border:SetWidth(edgeSize*8)
		border:SetHeight(edgeSize)
		border:SetPoint("BOTTOMLEFT", frame, "BOTTOMLEFT", edgeSize - 5, -5)
		border:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT", -edgeSize + 5, -5)
		border:SetTexCoord(0.375, 9.2808, 0.5, 9.2808, 0.375, 0, 0.5, 0)
		local border = newFrame("Texture", frame, "BORDER")
		frame.border5 = border
		border:SetTexture(edgeFile)
		border:SetVertexColor(unpack(self.db.profile.settings.colors.frameBorder))
		border:SetWidth(edgeSize)
		border:SetHeight(edgeSize)
		border:SetPoint("TOPLEFT", frame, "TOPLEFT", -5, 5)
		border:SetTexCoord(0.5, 0, 0.5, 1, 0.625, 0, 0.625, 1)
		local border = newFrame("Texture", frame, "BORDER")
		frame.border6 = border
		border:SetTexture(edgeFile)
		border:SetVertexColor(unpack(self.db.profile.settings.colors.frameBorder))
		border:SetWidth(edgeSize)
		border:SetHeight(edgeSize)
		border:SetPoint("TOPRIGHT", frame, "TOPRIGHT", 5, 5)
		border:SetTexCoord(0.625, 0, 0.625, 1, 0.75, 0, 0.75, 1)
		local border = newFrame("Texture", frame, "BORDER")
		frame.border7 = border
		border:SetTexture(edgeFile)
		border:SetVertexColor(unpack(self.db.profile.settings.colors.frameBorder))
		border:SetWidth(edgeSize)
		border:SetHeight(edgeSize)
		border:SetPoint("BOTTOMLEFT", frame, "BOTTOMLEFT", -5, -5)
		border:SetTexCoord(0.75, 0, 0.75, 1, 0.875, 0, 0.875, 1)
		local border = newFrame("Texture", frame, "BORDER")
		frame.border8 = border
		border:SetTexture(edgeFile)
		border:SetVertexColor(unpack(self.db.profile.settings.colors.frameBorder))
		border:SetWidth(edgeSize)
		border:SetHeight(edgeSize)
		border:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT", 5, -5)
		border:SetTexCoord(0.875, 0, 0.875, 1, 1, 0, 1, 1)
	end
	--[[
	local bg = newHash(
		'tile', true,
		'tileSize', 16,
		'bgFile', "Interface\\ChatFrame\\ChatFrameBackground",
		'edgeFile', edgeFile,
		'edgeSize', edgeSize,
		'insets', newHash(
			'left', inset,
			'right', inset,
			'top', inset,
			'bottom', inset
		)
	)
	frame:SetBackdrop(bg)
	bg.insets = del(bg.insets)
	bg = del(bg)
	frame:SetBackdropColor(unpack(self.db.profile.settings.colors.frameBG))
	frame:SetBackdropBorderColor(unpack(self.db.profile.settings.colors.frameBorder))
	]]
end

function PitBull:UpdateAllBackdropsAndBorders()
	for unit, frame in self:IterateUnitFrames() do
		self:UpdateBackdropAndBorder(frame)
	end
end

function PitBull:AreUnitFramesClampedToScreen()
	return self.db.profile.settings.clamped
end

function PitBull:ToggleUnitFramesClampedToScreen(value)
	if value == nil then
		value = not self.db.profile.settings.clamped
	end
	self.db.profile.settings.clamped = value
	for frame in pairs(allFrames) do
		frame:SetClampedToScreen(value)
	end
end

function PitBull:PopulateUnitFrame(frame)
	if not isframe(frame) then
		error(("Bad argument #2 to `PopulateUnitFrame'. Expected %s, got %s"):format("frame", type(frame)), 2)
	end
	if frames[frame] then
		self:UpdateFrame(frame)
		return
	else
		if not allFrames[frame] then
			return
		end
	end
	local unit = frame:GetUnit()
	
	local group = frame.group
	self:AddGroupToAceOptions(group)
	if not unit or self.db.profile.groups[group].hidden or (not UnitExists(unit) and not configMode) then
		self:ClearUnitFrame(frame)
		return
	end
	
	local overlay = newFrame("Frame", frame)
	frame.overlay = overlay
	overlay:SetFrameLevel(frame:GetFrameLevel()+3)
	
	for _, module in ipairs(modulesWithFunction.OnPopulateUnitFrame) do
		module:OnPopulateUnitFrame(unit, frame)
	end
	
	frames[frame] = true
	framesByUnit[unit] = frame
	if not framesByGroup[group] then
		framesByGroup[group] = new()
	end
	framesByGroup[group][frame] = true
	if not IsWackyGroup[group] then
		nonWackyFrames[unit] = frame
	end
	
	self:UpdateLayout(frame)
	self:UpdateBackdropAndBorder(frame)
	self:UpdateFrame(frame)
end

function PitBull:ClearUnitFrame(frame)
	if not isframe(frame) then
		error(("Bad argument #2 to `ClearUnitFrame'. Expected %s, got %s"):format("frame", type(frame)), 2)
	end
	frame:StopMovingOrSizing()
	frame:GetParent():StopMovingOrSizing()
	if not frames[frame] then
		return
	end
	local unit = frame:GetUnit()
	
	local group = frame.group
	
	local current_layout = layouts[self.db.profile.groups[frame.group].layout] or layouts.ABF
	
	if current_layout.extraFrames then
		for _, v in ipairs(current_layout.extraFrames) do
			if frame[v] then
				frame[v] = delFrame(frame[v])
			end
		end
	end
	
	for _, module in ipairs(modulesWithFunction.OnClearUnitFrame) do
		module:OnClearUnitFrame(unit, frame)
	end
	
	frame.overlay = delFrame(frame.overlay)
	
	if frame.backdrop then
		frame.backdrop = delFrame(frame.backdrop)
	end
	if frame.border1 then
		for i = 1, 8 do
			frame['border' .. i] = delFrame(frame['border' .. i])
		end
	end
	
	if not unit then
		unit = 'dummy'
	end
	
	frames[frame] = nil
	framesByUnit[unit] = nil
	framesByGroup[group][frame] = nil
	if not next(framesByGroup[group]) then
		framesByGroup[group] = del(framesByGroup[group])
	end
	nonWackyFrames[unit] = nil
end

local function frame_custom_menu(this, unit)
	Dewdrop:Open(this,
		'children', PitBull.options.args[this.group],
		'cursorX', true,
		'cursorY', true
	)
end

local function frame_menu(this, unit)
	local type, num
	
	if unit == "player" then
		type = PlayerFrameDropDown
	elseif unit == "target" then
		type = TargetFrameDropDown
	elseif unit == "pet" then
		type = PetFrameDropDown
	else
		num = unit:match("^party(%d)$")
		if num then
			type = _G["PartyMemberFrame" .. num .. "DropDown"]
		else
			num = unit:match("^raid(%d%d?)$")
			if num then
				type = FriendsDropDown
				this.unit = unit
				this.name = UnitName(unit)
				this.id = this:GetID()
				FriendsDropDown.displayMode = "MENU"
				FriendsDropDown.initialize = RaidFrameDropDown_Initialize
			else
				return frame_custom_menu(this, unit)
			end
		end
	end
	
	if num then
		PitBull:ScheduleLeaveCombatAction(this, "SetID", tonumber(num))
	end

	if type then
		HideDropDownMenu(1)
		type.unit = unit
		type.name = UnitName(unit)
		ToggleDropDownMenu(1, nil, type, "cursor")
		return true
	end
end

local function frame_OnEnter(this)
	if not frames[this] then
		return
	end
	local unit = this:GetUnit()
	for _, module in ipairs(modulesWithFunction.OnFrameOnEnter) do
		module:OnFrameOnEnter(unit, this)
	end
	if not PitBull.db.profile.settings.hideTooltipInCombat or not InCombatLockdown() then
		this.unit = unit
		UnitFrame_OnEnter()
		this.unit = nil
	end
end

local function frame_OnLeave(this)
	if not frames[this] then
		return
	end
	local unit = this:GetUnit()
	for _, module in ipairs(modulesWithFunction.OnFrameOnLeave) do
		module:OnFrameOnLeave(unit, this)
	end
	UnitFrame_OnLeave()
end

local function frame_OnDragStart(this)
	if PitBull.db.profile.settings.locked then
		return
	end
	if not frames[this] then
		return
	end
	-- also want to move group if grouped
	local cluster = this.cluster
	if cluster --[[ and not PitBull.db.profile.groups[this.group].freeform ]] then
		clusters[cluster]:StartMoving()
	else
		this:StartMoving()
	end
end

local function frame_OnDragStop(this)
	if not frames[this] then
		return
	end
	
	local cluster = this.cluster
	local freeform = false --PitBull.db.profile.groups[this.group].freeform
	local db
	if cluster and not freeform then
		this = clusters[cluster]
		db = self.db.profile.clusters[cluster]
	else
		local unit = this:GetUnit()
		db = self.db.profile.units[unit]
	end
	
	this:StopMovingOrSizing()
	
	local x, y = this:GetCenter()
	local s = this:GetEffectiveScale()
	local uis = UIParent:GetScale()
	this:ClearAllPoints()
	local point
	if cluster and not freeform then
		local direction = self.db.profile.groups[this.group].direction
		if direction == "up" then
			point = "BOTTOM"
			y = this:GetBottom()
		elseif direction == "right" then
			point = "LEFT"
			x = this:GetLeft()
		elseif direction == "left" then
			point = "RIGHT"
			x = this:GetRight()
		else
			point = "TOP"
			y = this:GetTop()
		end
	else
		point = "CENTER"
	end
	x = x*s - GetScreenWidth()*uis/2
	y = y*s - GetScreenHeight()*uis/2
	db.x, db.y = x/uis, y/uis
	this:SetPoint(point, UIParent, "CENTER", x/s, y/s)
	
	if cluster and not freeform then
		PitBull:ReorganizeCluster(cluster)
	end
end

local function frame_OnShow(this)
	local unit = this:GetUnit()
	if not unit then
		return
	end
	
	if not ShouldShowUnit(unit) then
		return
	elseif PitBull.db.profile.groups[this.group].hidden then
		return
	end
	PitBull:PopulateUnitFrame(this)
end

local function frame_OnHide(this)
	PitBull:ClearUnitFrame(this)
end

local function UnitFrame_GetUnit(self)
	local unit = self:GetAttribute("unit")
	if unit then
		local suffix = self:GetAttribute("unitsuffix")
		if suffix then
			unit = unit .. suffix
			unit = unit:gsub("^([^%d]+)([%d]+)pet", "%1pet%2")
		end
		return unit
	else
		local group = self.group
		if group then
			if group:find("^raidpet") then
				return "raidpet" .. (NumRaidMembers+1) .. group:sub(8)
			elseif group:find("^raid") then
				return "raid" .. (NumRaidMembers+1) .. group:sub(5)
			elseif group:find("^partypet") then
				return "partypet" .. (NumPartyMembers+1) .. group:sub(9)
			elseif group:find("^party") then
				return "party" .. (NumPartyMembers+1) .. group:sub(6)
			else
				return group
			end
		end
		return nil
	end
end

local count = 0
local cache = {}
local function configFrame_OnMouseDown(this, button)
	local unit = this:GetUnit()
	if not UnitExists(unit) and button == "RightButton" then
		this:custom_menu(unit)
	end
end
function PitBull:CreateUnitFrame(unit, secure)
	if type(unit) == "table" then
		preframe = unit
		unit = nil
	elseif type(unit) ~= "string" then
		error(("Bad argument #2 to `CreateUnitFrame'. Expected %s, got %s"):format("string or table", type(unit)), 2)
	end
	local frame
	if not preframe then
		local frame = framesByUnit[unit]
		if frame then
			self:UpdateFrame(frame)
			return frame
		end
	elseif frames[preframe] then
		self:UpdateFrame(preframe)
		return preframe
	end
	
	if not preframe and not IsLegitimateUnit[unit] then
		error(("Bad argument #2 to `CreateUnitFrame': %q is not a legitimate unit"):format(unit), 2)
	end
	
	local setup = false
	if not preframe then
		if #cache == 0 then
			count = count + 1
			local name = "PitBullUnitFrame" .. count
			frame = CreateFrame("Button", name, UIParent, "SecureUnitButtonTemplate")
			frame:Hide()
			setup = true
		else
			frame = table_remove(cache)
		end
	else
		frame = preframe
		if frame:GetScript("OnEnter") ~= frame_OnEnter then
			setup = true
		end
	end
	
	if setup then
		frame.GetUnit = UnitFrame_GetUnit
		frame:EnableMouse(true)
		frame:SetScript("OnEnter", frame_OnEnter)
		frame:SetScript("OnLeave", frame_OnLeave)
		frame:SetScript("OnDragStart", frame_OnDragStart)
		frame:SetScript("OnDragStop", frame_OnDragStop)
		frame:SetScript("OnHide", frame_OnHide)
		frame:SetScript("OnShow", frame_OnShow)
		frame:SetMovable(true)
		frame:RegisterForDrag("LeftButton")
		frame:RegisterForClicks("LeftButtonUp","RightButtonUp","MiddleButtonUp","Button4Up","Button5Up")
		frame:SetAttribute("*type1", "target")
		frame:SetAttribute("type2", "menu")
		frame:SetAttribute("*type2", "custom_menu")
		frame.menu = frame_menu
		frame.custom_menu = frame_custom_menu
		frame:SetClampedToScreen(self.db.profile.settings.clamped)
	end
	
	local group, cluster
	if not preframe then
		group = unit:gsub("%d", "")
		frame.group = group
	else
		group = frame.group
		cluster = frame.cluster
	end
	
	allFrames[frame] = true
	if not allFramesByGroup[group] then
		allFramesByGroup[group] = new()
	end
	allFramesByGroup[group][frame] = true
	if not preframe then
		allFramesByUnit[unit] = frame
	end
	if secure then
		frame:SetAttribute("initial-width", self.db.profile.groups[group].width)
		frame:SetAttribute("initial-height", self.db.profile.groups[group].height)
		frame:SetAttribute("initial-scale", self.db.profile.groups[group].scale)
	else
		frame:SetWidth(self.db.profile.groups[group].width)
		frame:SetHeight(self.db.profile.groups[group].height)
		frame:SetScale(self.db.profile.groups[group].scale)
	end

	-- db.{x,y} are scaled coordinates, in the UIParent scale.  Switch from
	-- that scale to the frame's scale.
	if cluster then
		if not allFramesByCluster[clusters[cluster]] then
			allFramesByCluster[clusters[cluster]] = new()
		end
		allFramesByCluster[clusters[cluster]][frame] = true
	else
		if not preframe then
			local db = self.db.profile.units[unit]
			local uis = UIParent:GetScale()
			local s = frame:GetEffectiveScale()
			frame:SetPoint("CENTER", db.x*uis/s, db.y*uis/s)
		end
	end
	
	if IsWackyGroup[group] then
		wackyFrames[frame] = true
	end
	
	if not preframe then
		frame:SetAttribute("unit", unit)
		
		if configMode or UnitExists(unit) then
			self:PopulateUnitFrame(frame)
		end
	end
	
	ClickCastFrames[frame] = true
	
	if not secure then
		if not configMode then
			RegisterUnitWatch(frame)
		else
			frame:Show()
			frame:SetScript("OnMouseDown", configFrame_OnMouseDown)
		end
	else
		frame:SetAttribute("initial-unitWatch", true)
	end
	
	return frame
end

function PitBull:DestroyUnitFrame(frame)
	if not isframe(frame) then
		error(("Bad argument #2 to `DestroyUnitFrame'. Expected %s, got %s"):format("frame", type(frame)), 2)
	end
	local unit = frame:GetUnit() or 'dummy'
	self:ClearUnitFrame(frame)
	cache[#cache+1] = frame
	UnregisterUnitWatch(frame)
	frame:Hide()
	frame:SetAttribute("unit", nil)
	local group = frame.group
	frame.group = nil
	frames[frame] = nil
	framesByUnit[unit] = nil
	allFrames[frame] = nil
	allFramesByUnit[unit] = nil
	if framesByGroup[group] then
		framesByGroup[group][frame] = nil
		if not next(framesByGroup[group]) then
			framesByGroup[group] = del(framesByGroup[group])
		end
	end
	allFramesByGroup[group][frame] = nil
	if not next(allFramesByGroup[group]) then
		allFramesByGroup[group] = del(allFramesByGroup[group])
	end
	if frame.cluster then
		local cluster = frame.cluster
		frame.cluster = nil
		allFramesByCluster[clusters[cluster]][frame] = nil
		if not next(allFramesByCluster[clusters[cluster]]) then
			allFramesByCluster[clusters[cluster]] = del(allFramesByCluster[clusters[cluster]])
		end
	end
	nonWackyFrames[unit] = nil
	if IsWackyGroup[group] then
		for frame in pairs(wackyFrames) do
			if frame:GetUnit() == unit then
				wackyFrames[frame] = nil
				break
			end
		end
	end
end

local count = 0
local partyCache = {}
local raidCache = {}
local function GetClusterGroupSuffixKind(name)
	local base = name:match("^(raid)") or name:match("^(party)") or nil
	local suffix = name:match("^raid(.+)$") or name:match("^party(.+)$") or nil
	local clusterKind
	if suffix then
		local goodsuffix = ''
		while suffix:find("target$") do
			suffix = suffix:sub(1, -7)
			goodsuffix = "target" .. goodsuffix
		end
		if suffix:find("pet$") then
			suffix = suffix:sub(1, -4)
			goodsuffix = "pet" .. goodsuffix
		end
		clusterKind = suffix
		suffix = goodsuffix
	end
	return name, suffix and (base .. suffix) or base, suffix, clusterKind
end

local classes = {"WARRIOR", "HUNTER", "ROGUE", "PALADIN", "SHAMAN", "PRIEST", "MAGE", "WARLOCK", "DRUID"}
function PitBull:CreateUnitCluster(cluster)
	if type(cluster) ~= "string" then
		error(("Bad argument #2 to `CreateUnitCluster'. Expected %s, got %s"):format("string", type(cluster)), 2)
	end
	if cluster == "raid" or cluster == "raidpet" or cluster == "raidtarget" or cluster == "raidpettarget" then
		if self.db.profile.groups.raid.groupStyle == "class" then
			for i, v in ipairs(classes) do
				self:CreateUnitCluster("raid" .. v .. cluster:sub(5))
			end
		else
			for i = 1, 8 do
				self:CreateUnitCluster("raid" .. i .. cluster:sub(5))
			end
		end
		return
	end
	if clusters[cluster] then
		return
	end
	local baseGroup
	if cluster:find("^raid") then
		baseGroup = "raid"
	elseif cluster:find("^party") then
		baseGroup = "party"
	else
		error(("Bad argument #3 to `CreateUnitCluster'. Expected %s, got %q"):format('"^party" or "^raid"', group), 2)
	end
	local cluster, group, suffix, clusterKind = GetClusterGroupSuffixKind(cluster)
	local header
	if baseGroup == "raid" and #raidCache > 0 then
		header = table_remove(raidCache)
	elseif baseGroup == "party" and #partyCache > 0 then
		header = table_remove(partyCache)
	else
		count = count + 1
		header = CreateFrame("Frame", "PitBullCluster" .. count, UIParent, baseGroup == "raid" and "SecureRaidGroupHeaderTemplate" or "SecurePartyHeaderTemplate")
	end
	clusters[cluster] = header
	header:SetMovable(true)
	header.group = group
	header.cluster = cluster
	local db = self.db.profile.clusters[cluster]
	header.initialConfigFunction = function(child)
		child.group = group
		child.cluster = cluster
		child:SetAttribute("unitsuffix", suffix)
		PitBull:CreateUnitFrame(child, true)
	end
	header:SetAttribute("template", "SecureUnitButtonTemplate")
	if self.db.profile.groups[baseGroup].nameSort then
		header:SetAttribute("sortMethod", "NAME")
	else
		header:SetAttribute("sortMethod", nil)
	end
	if baseGroup == "raid" then
		header:SetAttribute("groupBy", nil)
		if clusterKind then
			header:SetAttribute("groupFilter", tostring(clusterKind))
		else
			header:SetAttribute("groupFilter", "1,2,3,4,5,6,7,8")
		end
	end
	self:ReorganizeCluster(cluster)
	for i = 1, 40 do
		local child = header:GetAttribute('child' .. i)
		if child then
			child.group = group
			child.cluster = cluster
			child:SetAttribute("unitsuffix", suffix)
			PitBull:CreateUnitFrame(child)
		else
			break
		end
	end
	header:Show()
	
	if configMode then
		for i = 1, 40 do
			local frame = header:GetAttribute('child' .. i)
			if frame then
				UnregisterUnitWatch(frame)
				frame:Show()
				frame:SetScript("OnMouseDown", configFrame_OnMouseDown)
			else
				break
			end
		end
	end
end

function PitBull:DestroyUnitCluster(cluster)
	if type(cluster) ~= "string" then
		error(("Bad argument #2 to `DestroyUnitCluster'. Expected %s, got %s"):format("string", type(frame)), 2)
	end
	local header = clusters[cluster]
	if not header then
		return
	end
	for i = 1, 40 do
		local child = header:GetAttribute("child" .. i)
		if child then
			self:ClearUnitFrame(child)
			child.group = nil
			child.cluster = nil
			child:SetAttribute("unit", nil)
			child:SetAttribute("unitsuffix", nil)
		else
			break
		end
	end
	if header.group:find("^party") then
		partyCache[#partyCache+1] = header
	else
		raidCache[#raidCache+1] = header
	end
	header:StopMovingOrSizing()
	header:Hide()
	header:SetAttribute("template", nil)
	header:SetAttribute("sortMethod", nil)
	header:SetAttribute("groupBy", nil)
	header:SetAttribute("groupFilter", nil)
	clusters[cluster] = nil
end

function PitBull:ChangeScale(group, value)
	if self.db.profile.groups[group].scale == value then
		return
	end
	self.db.profile.groups[group].scale = value
	
	if allFramesByGroup[group] then
		for frame in pairs(allFramesByGroup[group]) do
			frame:SetScale(value)
			local cluster = frame.cluster
			
			if frames[frame] then
				self:UpdateLayout(frame)
			end
		end
	end
end

function PitBull:ChangeWidth(group, value)
	if self.db.profile.groups[group].width == value then
		return
	end
	self.db.profile.groups[group].width = value
	
	if allFramesByGroup[group] then
		for frame in pairs(allFramesByGroup[group]) do
			frame:SetWidth(value)
			
			if frames[frame] then
				self:UpdateLayout(frame)
			end
		end
	end
end

function PitBull:ChangeHeight(group, value)
	if self.db.profile.groups[group].height == value then
		return
	end
	self.db.profile.groups[group].height = value
	
	if allFramesByGroup[group] then
		for frame in pairs(allFramesByGroup[group]) do
			frame:SetHeight(value)
			
			if frames[frame] then
				self:UpdateLayout(frame)
			end
		end
	end
end

function PitBull:GetStatusBarTexture()
	return SharedMedia:Fetch('statusbar', self.db.profile.settings.texture)
end

function PitBull:GetFont()
	return SharedMedia:Fetch('font', self.db.profile.settings.font), self.db.profile.settings.fontsize
end

function PitBull:UpdateAllStatusBarTextures()
	local texture = PitBull:GetStatusBarTexture()
	for _,module in ipairs(modulesWithFunction.OnUpdateStatusBarTexture) do
		module:OnUpdateStatusBarTexture(texture)
	end
end

function PitBull:UpdateAllFonts()
	local font, fontsize = PitBull:GetFont()
	for _,module in ipairs(modulesWithFunction.OnUpdateFont) do
		module:OnUpdateFont(font, fontsize)
	end
end

function PitBull:SharedMedia_Registered(kind, name)
	if kind == "statusbar" then
		if self.db.profile and name == self.db.profile.settings.texture then
			self:UpdateAllStatusBarTextures()
		end
	elseif kind == "font" then
		if self.db.profile and name == self.db.profile.settings.font then
			self:UpdateAllFonts()
		end
	end
end

function PitBull:GetUnitFrame(unit)
	return framesByUnit[unit]
end

local function iter(tmp)
	local n = tmp.n + 1
	tmp.n = n
	local frame = tmp[n]
	if not frame then
		tmp = del(tmp)
		return
	end
	tmp[n] = nil
	return frame:GetUnit(), frame
end
function PitBull:IterateUnitFrames()
	local tmp = new()
	for frame in pairs(frames) do
		tmp[#tmp+1] = frame
	end
	tmp.n = 0
	return iter, tmp
end

local function blankIter() end
function PitBull:IterateUnitFramesByGroup(group)
	local g = framesByGroup[group]
	if not g then
		return blankIter
	else
		local tmp = new()
		for frame in pairs(g) do
			tmp[#tmp+1] = frame
		end
		tmp.n = 0
		return iter, tmp
	end
end

function PitBull:IterateNonWackyUnitFrames()
	return pairs(nonWackyFrames)
end

function PitBull:OnProfileEnable()
	self.colorConstants = self.db.profile.settings.colors
	self:UpdateAllStatusBarTextures()
	for _,module in ipairs(modulesWithFunction.OnPitBullProfileEnable) do
		module:OnPitBullProfileEnable()
	end
	
	self:UpdateAll()
end

function PitBull:RegisterLayout(name, data)
	layouts[name] = data
	layoutNames[name] = name
	
	for unit, frame in self:IterateUnitFrames() do
		if self.db.profile.groups[frame.group].layout == name then
			self:UpdateLayout(frame)
		end
	end
end

function PitBull:RegisterBorder(name, texture, size, insets)
	borders[name] = { texture, size, insets }
	borderNames[name] = name
	
	for unit, frame in self:IterateUnitFrames() do
		if self.db.profile.groups[frame.group].border == name then
			self:UpdateBackdropAndBorder(frame)
		end
	end
end

function PitBull:UpdateLayoutSettings(group)
	local layout = layouts[self.db.profile.groups[group].layout] or layouts.ABF
	
	for k, v in pairs(layout) do
		if type(k) == "string" and type(v) == "function" then
			local name = k:match("^(.+)_options$")
			if name and not self.options.args[group].args.layoutSettings.args[name] then
				local options = v(group)
				self.options.args[group].args.layoutSettings.args[name] = options
				if not options.handler then
					options.handler = layout
				end
				if name ~= "always" then
					local function func()
						local good = false
						for unit, frame in self:IterateUnitFramesByGroup(group) do
							if frame[name] then
								good = true
								break
							end
						end
						return not good
					end
					if options.hidden then
						if type(options.hidden) == "function" then
							local options_hidden = options.hidden
							function options.hidden(...)
								if options_hidden(...) then
									return true
								end
								return func()
							end
						elseif type(options.hidden) == "string" then
							local options_hidden = options.hidden
							function options.hidden(...)
								local handler = options.handler
								if handler[options_hidden](handler, ...) then
									return true
								end
								return func()
							end
						end
					else
						options.hidden = func
					end
				end
			end
		end
	end
end

function PitBull:ChangeLayout(group, value)
	local old_layout = layouts[self.db.profile.groups[group].layout] or layouts.ABF
	local new_layout = layouts[value] or layouts.ABF
	
	self.db.profile.groups[group].layout = value
	if old_layout == new_layout then
		return
	end
	
	if old_layout.extraFrames then
		for unit, frame in self:IterateUnitFramesByGroup(group) do
			for _, v in ipairs(old_layout.extraFrames) do
				if frame[v] then
					frame[v] = delFrame(frame[v])
				end
			end
		end
	end
	for k, v in pairs(old_layout) do
		if type(k) == "string" and k:find("_options$") then
			self.options.args[group].args.layoutSettings.args[k] = nil
		end
	end
	
	for unit, frame in self:IterateUnitFramesByGroup(group) do
		self:UpdateLayout(frame)
	end
end

function PitBull:ChangeBorder(group, value)
	self.db.profile.groups[group].border = value
	
	for unit, frame in self:IterateUnitFramesByGroup(group) do
		self:UpdateBackdropAndBorder(frame)
	end
end

function PitBull:ChangeHidden(group, value)
	self.db.profile.groups[group].hidden = value
	
	if value then
		if allFramesByGroup[group] then
			local destroyClusters = new()
			for frame in pairs(allFramesByGroup[group]) do
				local cluster = frame.cluster
				if cluster then
					destroyClusters[cluster] = true
				else
					self:DestroyUnitFrame(frame)
				end
			end
			for cluster in pairs(destroyClusters) do
				self:DestroyUnitCluster(cluster)
			end
			destroyClusters = del(destroyClusters)
		end
	else
		if not ShouldShowUnit(group) then
			return
		end
		
		if group:find("^party") then
			if configMode == "party" or configMode == "raid" or NumPartyMembers > 0 then
				self:CreateUnitCluster(group)
			end
		elseif group:find("^raid") then
			if configMode == "raid" or NumRaidMembers > 0 then
				self:CreateUnitCluster(group)
			end
		else
			self:CreateUnitFrame(group)
		end
	end
end

function PitBull:ChangeGroupFreeform(group, value)
	self.db.profile.groups[group].freeform = value
	
	for cluster in pairs(clusters) do
		if cluster:find("^" .. group) then
			self:ReorganizeCluster(cluster)
		end
	end
end

function PitBull:ChangeGroupDirection(group, value)
	self.db.profile.groups[group].direction = value
	
	for cluster in pairs(clusters) do
		if cluster:find("^" .. group) then
			self:ReorganizeCluster(cluster)
		end
	end
end

function PitBull:ChangeGroupSpacing(group, value)
	self.db.profile.groups[group].spacing = value
	
	for cluster in pairs(clusters) do
		if cluster:find("^" .. group) then
			self:ReorganizeCluster(cluster)
		end
	end
end

function PitBull:ChangeHidePartyInRaid(value)
	self.db.profile.groups.party.hidePartyInRaid = value
	
	self:PARTY_MEMBERS_CHANGED()
	
	self:ChangeConfigMode(configMode)
end

function PitBull:ChangeShow5manAsParty(value)
	self.db.profile.groups.raid.show5manAsParty = value
	
	local config = configMode
	self:ChangeConfigMode(nil)
	
	self:PARTY_MEMBERS_CHANGED()

	self:ChangeConfigMode(config)
end

function PitBull:ChangeNameSort(group, value)
	self.db.profile.groups[group].nameSort = value
	
	for cluster, header in pairs(clusters) do
		local baseGroup = cluster:match("^(raid)") or cluster:match("^(party)") or nil
		if baseGroup then
			if self.db.profile.groups[baseGroup].nameSort then
				header:SetAttribute("sortMethod", "NAME")
			else
				header:SetAttribute("sortMethod", nil)
			end
		end
	end
end

function PitBull:ChangeRaidGroupStyle(value)
	for cluster in pairs(clusters) do
		if cluster:find("^raid") then
			self:DestroyUnitCluster(cluster)
		end
	end
	
	self.db.profile.groups.raid.groupStyle = value
	
	for group, db in pairs(self.db.profile.groups) do
		if not db.hidden and group:find("^raid") then
			self:CreateUnitCluster(group)
		end
	end
end

local mySecurePartyHeader_Update, mySecureRaidGroupHeader_Update
do
	-- relativePoint, xMultiplier, yMultiplier = getRelativePointAnchor( point )
	-- Given a point return the opposite point and which axes the point
	-- depends on.
	local function getRelativePointAnchor( point )
		point = strupper(point);
		if (point == "TOP") then
			return "BOTTOM", 0, -1;
		elseif (point == "BOTTOM") then
			return "TOP", 0, 1;
		elseif (point == "LEFT") then
			return "RIGHT", 1, 0;
		elseif (point == "RIGHT") then
			return "LEFT", -1, 0;
		elseif (point == "TOPLEFT") then
			return "BOTTOMRIGHT", 1, -1;
		elseif (point == "TOPRIGHT") then
			return "BOTTOMLEFT", -1, -1;
		elseif (point == "BOTTOMLEFT") then
			return "TOPRIGHT", 1, 1;
		elseif (point == "BOTTOMRIGHT") then
			return "TOPLEFT", -1, 1;
		else
			return "CENTER", 0, 0;
		end
	end
	
	-- empties tbl and assigns the value true to each key passed as part of ...
	local function fillTable( tbl, ... )
		for key in pairs(tbl) do
			tbl[key] = nil;
		end
		for i = 1, select("#", ...), 1 do
			local key = select(i, ...);
			key = tonumber(key) or key;
			tbl[key] = true;
		end
	end

	-- same as fillTable() except that each key is also stored in 
	-- the array portion of the table in order
	local function doubleFillTable( tbl, ... )
		fillTable(tbl, ...);
		for i = 1, select("#", ...), 1 do
			tbl[i] = select(i, ...);
		end
	end

	--working tables
	local tokenTable = {};
	local sortingTable = {};
	local groupingTable = {};
	local tempTable = {};
	
	local englishClasses = { "WARRIOR", "HUNTER", "ROGUE", "PALADIN", "SHAMAN", "PRIEST", "MAGE", "WARLOCK", "DRUID" }

	-- creates child frames and finished configuring them
	local function configureChildren(self)
		local point = self:GetAttribute("point") or "TOP"; --default anchor point of "TOP"
		local relativePoint, xOffsetMult, yOffsetMult = getRelativePointAnchor(point);
		local xMultiplier, yMultiplier =  abs(xOffsetMult), abs(yOffsetMult);
		local xOffset = self:GetAttribute("xOffset") or 0; --default of 0
		local yOffset = self:GetAttribute("yOffset") or 0; --default of 0
		local sortDir = self:GetAttribute("sortDir") or "ASC"; --sort ascending by default
		local columnSpacing = self:GetAttribute("columnSpacing") or 0;
		local startingIndex = self:GetAttribute("startingIndex") or 1;
	
		local unitCount = #sortingTable;
		local numDisplayed = unitCount - (startingIndex - 1);
		local unitsPerColumn = self:GetAttribute("unitsPerColumn");
		local numColumns;
		if ( unitsPerColumn and numDisplayed > unitsPerColumn ) then
			numColumns = min( ceil(numDisplayed / unitsPerColumn), (self:GetAttribute("maxColumns") or 1) );
		else
			unitsPerColumn = numDisplayed;
			numColumns = 1;
		end
		local loopStart = startingIndex;
		local loopFinish = min((startingIndex - 1) + unitsPerColumn * numColumns, unitCount)
		local step = 1;
	
		numDisplayed = loopFinish - (loopStart - 1);
	
		if ( sortDir == "DESC" ) then
			loopStart = unitCount - (startingIndex - 1);
			loopFinish = loopStart - (numDisplayed - 1);
			step = -1;
		end
	
		-- ensure there are enough buttons
		local needButtons = max(1, numDisplayed);
		if not ( self:GetAttribute("child"..needButtons) ) then
			local buttonTemplate = self:GetAttribute("template");
			local templateType = self:GetAttribute("templateType") or "Button";
			local name = self:GetName();
			if not ( buttonTemplate and name ) then
				self:Hide();
				return;
			end
			for i = 1, needButtons, 1 do
				if not ( self:GetAttribute("child"..i) ) then
					local newButton = CreateFrame(templateType, name.."UnitButton"..i, self, buttonTemplate);
					SetupUnitButtonConfiguration(self, newButton);
					self:SetAttribute("child"..i, newButton);
				end
			end
		end
	
		local columnAnchorPoint, columnRelPoint, colxMulti, colyMulti;
		if ( numColumns > 1 ) then
			columnAnchorPoint = self:GetAttribute("columnAnchorPoint");
			columnRelPoint, colxMulti, colyMulti = getRelativePointAnchor(columnAnchorPoint);
		end
	
		local buttonNum = 0;
		local columnNum = 1;
		local columnUnitCount = 0;
		local currentAnchor = self;
		for i = loopStart, loopFinish, step do
			buttonNum = buttonNum + 1;
			columnUnitCount = columnUnitCount + 1;
			if ( columnUnitCount > unitsPerColumn ) then
				columnUnitCount = 1;
				columnNum = columnNum + 1;
			end
		
			local unitButton = self:GetAttribute("child"..buttonNum);
			unitButton:Hide();
			unitButton:ClearAllPoints();
			if ( buttonNum == 1 ) then
				unitButton:SetPoint(point, currentAnchor, point, 0, 0);
				if ( columnAnchorPoint ) then
					unitButton:SetPoint(columnAnchorPoint, currentAnchor, columnAnchorPoint, 0, 0);
				end
		
			elseif ( columnUnitCount == 1 ) then
				local columnAnchor = self:GetAttribute("child"..(buttonNum - unitsPerColumn));
				unitButton:SetPoint(columnAnchorPoint, columnAnchor, columnRelPoint, colxMulti * columnSpacing, colyMulti * columnSpacing);
		
			else
				unitButton:SetPoint(point, currentAnchor, relativePoint, xMultiplier * xOffset, yMultiplier * yOffset);
			end
			unitButton:SetAttribute("unit", sortingTable[sortingTable[i]]);
			unitButton:Show();
		
			currentAnchor = unitButton;
		end
		repeat
			buttonNum = buttonNum + 1;
			local unitButton = self:GetAttribute("child"..buttonNum);
			if ( unitButton ) then
				unitButton:Hide();
				unitButton:SetAttribute("unit", nil);
			end
		until not ( unitButton )
	
		local unitButton = self:GetAttribute("child1");
		local unitButtonWidth = unitButton:GetWidth();
		local unitButtonHeight = unitButton:GetHeight();
		if ( numDisplayed > 0 ) then
			local width = xMultiplier * (unitsPerColumn - 1) * unitButtonWidth + ( (unitsPerColumn - 1) * (xOffset * xOffsetMult) ) + unitButtonWidth;
			local height = yMultiplier * (unitsPerColumn - 1) * unitButtonHeight + ( (unitsPerColumn - 1) * (yOffset * yOffsetMult) ) + unitButtonHeight;
		
			if ( numColumns > 1 ) then
				width = width + ( (numColumns -1) * abs(colxMulti) * (width + columnSpacing) );
				height = height + ( (numColumns -1) * abs(colyMulti) * (height + columnSpacing) );
			end
		
			self:SetWidth(width);
			self:SetHeight(height);
		else
			local minWidth = self:GetAttribute("minWidth") or (yMultiplier * unitButtonWidth);
			local minHeight = self:GetAttribute("minHeight") or (xMultiplier * unitButtonHeight);
			self:SetWidth( max(minWidth, 0.1) );
			self:SetHeight( max(minHeight, 0.1) );
		end
	end

	function mySecurePartyHeader_Update(self)
		if not configMode or configMode == "solo" then
			return
		end
		for key in pairs(sortingTable) do
			sortingTable[key] = nil;
		end

		for i = 1, GetNumPartyMembers(), 1 do
			local unit = "party"..i;
			local name = UnitName(unit);
			tinsert(sortingTable, name);
			sortingTable[name] = unit;
		end
	
		if ( self:GetAttribute("sortMethod") == "NAME" ) then -- sort by ID by default
			table.sort(sortingTable);
		end
		
		-- my addition
		for i = GetNumPartyMembers()+1, 4 do
			local unit = "party"..i;
			tinsert(sortingTable, unit);
			sortingTable[unit] = unit;
		end
		-- end my addition

		configureChildren(self);
	end
	
	function mySecureRaidGroupHeader_Update(self)
		if configMode ~= "raid" then
			return
		end
		local nameList = self:GetAttribute("nameList");
		local groupFilter = self:GetAttribute("groupFilter");
		local sortMethod = self:GetAttribute("sortMethod");
		local groupBy = self:GetAttribute("groupBy");

		if not ( groupFilter or nameList ) then
			self:Hide();
			return;
		end

		for key in pairs(sortingTable) do
			sortingTable[key] = nil;
		end

		if ( groupFilter ) then
			-- filtering by a list of group numbers and/or classes
			fillTable(tokenTable, strsplit(",", groupFilter));
			local strictFiltering = self:GetAttribute("strictFiltering"); -- non-strict by default
			for i = 1, GetNumRaidMembers(), 1 do
				local name, _, subgroup, _, _, className = GetRaidRosterInfo(i);
				if ( ((not strictFiltering) and 
				      (tokenTable[subgroup] or tokenTable[className]) -- non-strict filtering
				     ) or 
				      (tokenTable[subgroup] and tokenTable[className]) -- strict filtering
				) then
					tinsert(sortingTable, name);
					sortingTable[name] = "raid"..i;
					if ( groupBy == "GROUP" ) then
						groupingTable[name] = subgroup;

					elseif ( groupBy == "CLASS" ) then
						groupingTable[name] = className;

					end
				end
			end

			if ( groupBy ) then
				local groupingOrder = self:GetAttribute("groupingOrder");
				doubleFillTable(tokenTable, strsplit(",", groupingOrder));
				for k in pairs(tempTable) do
					tempTable[k] = nil;
				end
				for _, grouping in ipairs(tokenTable) do
					grouping = tonumber(grouping) or grouping;
					for k in ipairs(groupingTable) do
						groupingTable[k] = nil;
					end
					for index, name in ipairs(sortingTable) do
						if ( groupingTable[name] == grouping ) then
							tinsert(groupingTable, name);
							tempTable[name] = true;
						end
					end
					if ( sortMethod == "NAME" ) then -- sort by ID by default
						table.sort(groupingTable);
					end
					for _, name in ipairs(groupingTable) do
						tinsert(tempTable, name);
					end
				end
				-- handle units whose group didn't appear in groupingOrder
				for k in ipairs(groupingTable) do
					groupingTable[k] = nil;
				end
				for index, name in ipairs(sortingTable) do
					if not ( tempTable[name] ) then
						tinsert(groupingTable, name);
					end
				end
				if ( sortMethod == "NAME" ) then -- sort by ID by default
					table.sort(groupingTable);
				end
				for _, name in ipairs(groupingTable) do
					tinsert(tempTable, name);
				end

				--copy the names back to sortingTable
				for index, name in ipairs(tempTable) do
					sortingTable[index] = name;
				end

			elseif ( sortMethod == "NAME" ) then -- sort by ID by default
				table.sort(sortingTable);

			end
			
			-- my addition
			local current = GetNumRaidMembers();
			local groups = new()
			for i = 1, 8 do
				if tokenTable[i] then
					for j = 1, current do
						local name, _, subgroup, _, _, className = GetRaidRosterInfo(j);
						groups[subgroup] = (groups[subgroup] or 0) + 1;
					end
					local num = current;
					for j = 1, i do
						num = num + 5 - (groups[j] or 0);
					end
					for j = (groups[i] or 0)+1, 5 do
						local unit = "raid"..(num+j-5);
						tinsert(sortingTable, unit);
						sortingTable[unit] = unit;
					end
				end
			end
			groups = del(groups)
			groups = new()
			for i,v in ipairs(englishClasses) do
				if tokenTable[v] then
					local num = current
					for j = 1, i-1 do
						local q = 40 - current
						if q%9 ~= 0 then
							if q%9 >= j then
								q = floor(q/9) + 1
							else
								q = floor(q/9)
							end
						else
							q = q/9
						end
						num = num + q
					end
					local q = 40 - current
					if q%9 ~= 0 then
						if q%9 >= i then
							q = floor(q/9) + 1
						else
							q = floor(q/9)
						end
					else
						q = q/9
					end
					for j = num+1, num+q do
						local unit = "raid"..j;
						tinsert(sortingTable, unit);
						sortingTable[unit] = unit;
					end
				end
			end
			groups = del(groups)
			-- end my addition

		elseif ( nameList ) then
			-- filtering via a list of names
			doubleFillTable(sortingTable, strsplit(",", nameList));
			for i = 1, GetNumRaidMembers(), 1 do
				local name = GetRaidRosterInfo(i);
				if ( sortingTable[name] ) then
					sortingTable[name] = "raid"..i;
				end
			end
			local current = GetNumRaidMembers() -- mine
			for i = #sortingTable, 1, -1 do
				local name = sortingTable[i];
				if ( sortingTable[name] == true ) then
					-- my addition
					current = current + 1
					sortingTable[name] = "raid"..current
					-- end my addition
				end
			end
			if ( sortMethod == "NAME" ) then
				table.sort(sortingTable);
			end

		else
			self:Hide();
			return;
		end

		configureChildren(self);
	end
end

local isPartyHooked, isRaidHooked = false, false

function PitBull:ChangeConfigMode(value)
	if value == "disabled" or not value then
		value = nil
	else
		if NumRaidMembers > 0 then
			if NumRaidMembers == NumPartyMembers+1 and self.db.profile.groups.raid.show5manAsParty then
				value = "party"
			else
				value = "raid"
			end
		elseif NumPartyMembers > 0 then
			value = "party"
		end
	end
	self.configMode = value
	configMode = value
	
	if configMode == "raid" then
		if not isPartyHooked then
			isPartyHooked = true
			hooksecurefunc("SecurePartyHeader_Update", mySecurePartyHeader_Update)
		end
		if not isRaidHooked then
			isRaidHooked = true
			hooksecurefunc("SecureRaidGroupHeader_Update", mySecureRaidGroupHeader_Update)
		end
		for group, db in pairs(self.db.profile.groups) do
			if not db.hidden and (group:find("^party") or group:find("^raid")) and ShouldShowUnit(group) then
				self:CreateUnitCluster(group)
			elseif group:find("^party") or group:find("^raid") then
				self:DestroyUnitCluster(group)
			end
		end
		for cluster, header in pairs(clusters) do
			if cluster:find("^party") then
				SecurePartyHeader_Update(header)
			else
				SecureRaidGroupHeader_Update(header)
			end
		end
	elseif configMode == "party" then
		if not isPartyHooked then
			isPartyHooked = true
			hooksecurefunc("SecurePartyHeader_Update", mySecurePartyHeader_Update)
		end
		for group, db in pairs(self.db.profile.groups) do
			if not db.hidden and group:find("^party") then
				self:CreateUnitCluster(group)
			elseif group:find("^party") or group:find("^raid") then
				self:DestroyUnitCluster(group)
			end
		end
		for cluster, header in pairs(clusters) do
			if cluster:find("^party") then
				SecurePartyHeader_Update(header)
			elseif cluster:find("^raid") then
				self:DestroyUnitCluster(cluster)
			end
		end
	end
	
	for frame in pairs(allFrames) do
		local unit = frame:GetUnit()
		if unit then
			if configMode then
				if not ShouldShowUnit(unit) then
					RegisterUnitWatch(frame)
					frame:SetScript("OnMouseDown", nil)
				else
					UnregisterUnitWatch(frame)
					frame:Show()
					frame:SetScript("OnMouseDown", configFrame_OnMouseDown)
				end
			else
				RegisterUnitWatch(frame)
				frame:SetScript("OnMouseDown", nil)
			end
		end
	end
	
	for _, module in ipairs(modulesWithFunction.OnChangeConfigMode) do
		module:OnChangeConfigMode(value)
	end
end

function PitBull.modulePrototype:RegisterPitBullChildFrames(...)
	for i = 1, select('#', ...) do
		local v = select(i, ...)
		if type(v) ~= "string" then
			error(("Bad argument #%d to `RegisterPitBullChildFrames'. Expected \"string\", got %q"):format(i, type(v)), 2)
		end
		if frameToModule[v] and frameToModule[v] ~= self then
			error(("Frame %q provided twice to `RegisterPitBullChildFrames'. Previously by %q and now by %q"):format(v, tostring(frameToModule[v]), tostring(self)), 2)
		end
		frameToModule[v] = self
	end
end

function PitBull.modulePrototype:RegisterPitBullOptionsMethod(method)
	local kind = type(method)
	if kind ~= "string" and kind ~= "function" then
		error(("Bad argument #2 to `RegisterPitBullOptionsMethod'. %s expected, got %s."):format("string or function", kind), 2)
	end
	if kind == "string" then
		if type(self[method]) ~= "function" then
			error(("Cannot find handler %q for `RegisterPitBullOptionsMethod'"):format(method), 2)
		end
	end
	
	if moduleOptionsMethods[self] then
		error("Cannot call `RegisterPitBullOptionsMethod' more than once from the same module.", 2)
	end
	
	local name = self.name
	
	for group in pairs(groupMenus) do
		local options = PitBull.options.args[group].args
		if kind == "function" then
			options[name] = method(group)
		else
			options[name] = self[method](self, group)
		end
		if options[name] and not options[name].handler then
			options[name].handler = self
		end
	end
	
	moduleOptionsMethods[self] = method
end

PitBull:RegisterBorder("Classic", "Interface\\Tooltips\\UI-Tooltip-Border", 16, 5)
PitBull:RegisterBorder("Nurfed", "Interface\\DialogFrame\\UI-DialogBox-Border", 16, 5)
PitBull:RegisterBorder("Hidden", "", 0, 3)

function SetupDebugging()
	local start = GetTime()
	local tree = {}
	local treeMemories = {}
	local treeTimes = {}
	local memories = {}
	local times = {}
	local calls = {}
	function wrap(value, name)
		if type(value) == "function" then
			local oldFunction = value
			memories[name] = 0
			times[name] = 0
			calls[name] = 0
			local function func(pos, t, mem, ...)
				mem, t = collectgarbage('count') - mem, GetTime() - t
				if pos > 0 then
					treeMemories[pos] = treeMemories[pos] + mem
					treeTimes[pos] = treeTimes[pos] + t
				end
				local otherMem = table.remove(treeMemories)
				if mem - otherMem > 0 then
					memories[name] = memories[name] + mem - otherMem
				end
				times[name] = times[name] + t - table.remove(treeTimes)
				table.remove(tree)
				return ...
			end
			return function(...)
				local pos = #tree
				tree[#tree+1] = name
				treeMemories[#treeMemories+1] = 0
				treeTimes[#treeTimes+1] = 0
				calls[name] = calls[name] + 1
				local t, mem = GetTime(), collectgarbage('count')
				return func(pos, t, mem, oldFunction(...))
			end
		elseif type(value) == "table" then
			for k, v in pairs(value) do
				if type(k) == "string" and type(v) == "function" and v ~= wrap then
					value[k] = wrap(v, name .. ':' .. k)
				end
			end
			return value
		end
	end

	wrap(PitBull, '')
	for name, module in PitBull:IterateModules() do
		wrap(module, name)
	end
	wrap(DogTag, 'DogTag-1.0')

	function ProfilePitBull()
		if not Tablet:IsRegistered("PitBull") then
			local currentSort = 'time'
			local sorts = {
				time = function(alpha, bravo)
					if times[alpha] ~= times[bravo] then
						return times[alpha] < times[bravo]
					elseif memories[alpha] ~= memories[bravo] then	
						return memories[alpha] < memories[bravo]
					elseif calls[alpha] ~= calls[bravo] then
						return calls[alpha] < calls[bravo]
					else
						return alpha:lower() < bravo:lower()
					end
				end,
				memory = function(alpha, bravo)
					if memories[alpha] ~= memories[bravo] then	
						return memories[alpha] < memories[bravo]
					elseif times[alpha] ~= times[bravo] then
						return times[alpha] < times[bravo]
					elseif calls[alpha] ~= calls[bravo] then
						return calls[alpha] < calls[bravo]
					else
						return alpha:lower() < bravo:lower()
					end
				end,
				calls = function(alpha, bravo)
					if calls[alpha] ~= calls[bravo] then
						return calls[alpha] < calls[bravo]
					elseif times[alpha] ~= times[bravo] then
						return times[alpha] < times[bravo]
					elseif memories[alpha] ~= memories[bravo] then	
						return memories[alpha] < memories[bravo]
					else
						return alpha:lower() < bravo:lower()
					end
				end,
				name = function(alpha, bravo)
					return alpha:lower() < bravo:lower()
				end,
			}
			Tablet:Register("PitBull",
				'cantAttach', true,
				'menu', {
					type = 'group',
					args = {
						time = {
							name = "Sort by time",
							desc = "Sort by time",
							type = 'toggle',
							get = function()
								return currentSort == 'time'
							end,
							set = function(value)
								currentSort = 'time'
							end,
						},
						memory = {
							name = "Sort by memory",
							desc = "Sort by memory",
							type = 'toggle',
							get = function()
								return currentSort == 'memory'
							end,
							set = function(value)
								currentSort = 'memory'
							end,
						},
						name = {
							name = "Sort by name",
							desc = "Sort by name",
							type = 'toggle',
							get = function()
								return currentSort == 'name'
							end,
							set = function(value)
								currentSort = 'name'
							end,
						},
						calls = {
							name = "Sort by calls",
							desc = "Sort by calls",
							type = 'toggle',
							get = function()
								return currentSort == 'calls'
							end,
							set = function(value)
								currentSort = 'calls'
							end,
						},
					},
				},
				'children', function()
					Tablet:SetTitle("PitBull Profile")
					
					local cat = Tablet:AddCategory(
						'columns', 7,
						'text', "Function",
						'text2', "Time",
						'text3', "Percent Time",
						'text4', "Memory",
						'text5', "Calls",
						'text6', "Time/Call",
						'text7', "Memory/Call"
					)
					
					local t = new()
					for method in pairs(memories) do
						t[#t+1] = method
					end
					table.sort(t, sorts[currentSort])
					local memory = 0
					local time = 0
					local calls_total = 0
					local duration = GetTime() - start
					for _,method in ipairs(t) do
						cat:AddLine(
							'text', method,
							'text2', ("%.3f s"):format(times[method]),
							'text3', ("%.3f%%"):format(times[method] / duration * 100),
							'text4', ("%.3f KiB"):format(memories[method]),
							'text5', ("%dx"):format(calls[method]),
							'text6', ("%.3f ms"):format(times[method] / calls[method] * 1000),
							'text7', ("%.3f KiB"):format(memories[method] / calls[method])
						)
						memory = memory + memories[method]
						time = time + times[method]
						calls_total = calls_total + calls[method]
					end
					cat:AddLine(
						'text', ' '
					)
					cat:AddLine(
						'text', "Total",
						'text2', ("%.3f s"):format(time),
						'text3', ("%.3f%%"):format(time / duration * 100),
						'text4', ("%.3f KiB"):format(memory),
						'text5', ("%dx"):format(calls_total),
						'text6', ("%.3f ms"):format(time / calls_total * 1000),
						'text7', ("%.3f KiB"):format(memory / calls_total)
					)
					del(t)
				end
			)
			ProfilePitBull_open = true
			PitBull:ScheduleRepeatingEvent("PitBull-UpdateProfileTablet", function()
				Tablet:Refresh("PitBull")
			end, 0.2)
			Tablet:Open("PitBull")
		else
			if ProfilePitBull_open then
				Tablet:Close("PitBull")
				ProfilePitBull_open = false
			else
				Tablet:Open("PitBull")
				ProfilePitBull_open = true
			end
		end
	end
end
