local VERSION = tonumber(("$Revision: 32674 $"):match("%d+"))

local PitBull = PitBull
local PitBull_Rest = PitBull:NewModule("Rest", "AceEvent-2.0")
local self = PitBull_Rest
if PitBull.revision < VERSION then
	PitBull.version = "r" .. VERSION
	PitBull.revision = VERSION
	PitBull.date = ("$Date: 2007-04-15 01:33:31 -0400 (Sun, 15 Apr 2007) $"):match("%d%d%d%d%-%d%d%-%d%d")
end

local new, del, newHash = PitBull.new, PitBull.del, PitBull.newHash
local newFrame, delFrame = PitBull.newFrame, PitBull.delFrame

PitBull_Rest:RegisterPitBullChildFrames('restIcon')

function PitBull_Rest:OnInitialize()
	self.db = PitBull:AcquireDBNamespace("Rest")
	PitBull:RegisterDefaults("Rest", "profile", {
		disable = false,
	})
end

function PitBull_Rest:OnEnable()
	self:RegisterEvent("PLAYER_UPDATE_RESTING")
	self:RegisterEvent("PLAYER_ENTERING_WORLD", "PLAYER_UPDATE_RESTING")
end

function PitBull_Rest:PLAYER_UPDATE_RESTING()
	local frame = PitBull:GetUnitFrame("player")
	if not frame then
		return
	end
	
	local resting = IsResting()
	if resting and not self.db.profile.disable then
		if not frame.restIcon then
			frame.restIcon = newFrame("Texture", frame.overlay, "ARTWORK")
			frame.restIcon:SetTexture("Interface\\CharacterFrame\\UI-StateIcon")
			frame.restIcon:SetTexCoord(0, 0.5, 0, 0.421875)
			
			PitBull:UpdateLayout(PitBull:GetUnitFrame("player"))
		end
	else
		if frame.restIcon then
			frame.restIcon = delFrame(frame.restIcon)
			
			PitBull:UpdateLayout(PitBull:GetUnitFrame("player"))
		end
	end
end

function PitBull_Rest:OnUpdateFrame(unit, frame)
	if unit == "player" then
		self:PLAYER_UPDATE_RESTING()
	end
end

function PitBull_Rest:OnClearUnitFrame(unit, frame)
	if frame.restIcon then
		frame.restIcon = delFrame(frame.restIcon)
	end
end

function PitBull_Rest:OnUnknownLayout(unit, frame)
	if frame.combatIcon then
		frame.restIcon:SetPoint("LEFT", frame.combatIcon, "RIGHT", 4, 0)
	else
		frame.restIcon:SetPoint("CENTER", frame, "BOTTOMLEFT", 0, 0)
	end
	frame.restIcon:SetWidth(18)
	frame.restIcon:SetHeight(18)
end

PitBull_Rest:RegisterPitBullOptionsMethod(function(group)
	if group == "player" then
		return {
			type = 'group',
			name = 'Rest icon',
			desc = 'Rest icon',
			args = {
				hidden = {
					type = 'toggle',
					name = "Hide rest icon",
					desc = "Hide rest icon",
					get = function()
						return self.db.profile.disable
					end,
					set = function(value)
						self.db.profile.disable = value
						self:PLAYER_UPDATE_RESTING()
					end,
				}
			}
		}
	end
end)